# Copyright 2020-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow: Gates peculiar to Cirq
"""

import numpy as np

from .. import tensors, var
from ..config import SWAP_TARGET
from ..ops import StdGate
from ..paulialgebra import Pauli
from ..qubits import Qubit
from ..tensors import QubitTensor
from ..utils import cached_property
from ..var import Variable
from .stdgates_1q import I, X, XPow, ZPow
from .stdgates_2q import CZ, Swap

__all__ = ("PhasedX", "PhasedXPow", "FSim", "FSwap", "FSwapPow", "Sycamore")


# Kudos: Phased X gate and powers of phased X gates adapted from Cirq


class PhasedX(StdGate):
    r"""A phased X gate, equivalent to the circuit
    ───Z^-p───X───Z^p───
    """
    # Kudos: Adapted from Cirq

    _diagram_labels = ["PhX({p})"]
    cv_tensor_structure = "monomial"

    def __init__(self, p: Variable, q0: Qubit) -> None:
        super().__init__(params=[p], qubits=[q0])

    @cached_property
    def tensor(self) -> QubitTensor:
        p = var.asfloat(self.param("p"))
        gate = ZPow(p, 0) @ X(0) @ ZPow(-p, 0)
        return gate.tensor

    @property
    def H(self) -> "PhasedX":
        return self

    def __pow__(self, t: Variable) -> "PhasedXPow":
        p = self.param("p")
        return PhasedXPow(p, t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        p = self.param("p") % 2
        if np.isclose(p, 0.0) or np.isclose(p, 2.0):
            return X(*qbs)
        return self


# end class PhasedX


class PhasedXPow(StdGate):
    """A phased X gate raised to a power.

    Equivalent to the circuit ───Z^-p───X^t───Z^p───

    """

    # Kudos: Adapted from Cirq

    _diagram_labels = ["PhX({p})^{t}"]

    def __init__(self, p: Variable, t: Variable, q0: Qubit) -> None:
        super().__init__(params=[p, t], qubits=[q0])

    @cached_property
    def tensor(self) -> QubitTensor:
        p, t = self.params
        gate = ZPow(p, 0) @ XPow(t, 0) @ ZPow(-p, 0)
        return gate.tensor

    @property
    def H(self) -> "PhasedXPow":
        return self ** -1

    def __pow__(self, t: Variable) -> "PhasedXPow":
        p, s = self.params
        return PhasedXPow(p, s * t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        p = self.param("p") % 2
        t = self.param("t") % 2
        if np.isclose(t, 0.0) or np.isclose(t, 2.0):
            return I(*qbs)
        if np.isclose(p, 0.0):
            return XPow(t, *qbs).specialize()
        return self


# end class PhasedXPow


class FSim(StdGate):
    r"""Fermionic simulation gate family.

    Contains all two qubit interactions that preserve excitations, up to
    single-qubit rotations and global phase.

    Locally equivalent to ``Can(theta/pi, theta/pi, phi/(2*pi))``

    .. math::
        \text{FSim}(\theta, \phi) =
            \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & \cos(\theta) & -i \sin(\theta) & 0 \\
                0 & -i sin(\theta)  & \cos(\theta) & 0 \\
                0 & 0 & 0 & e^{-i\phi)}
            \end{pmatrix}
    """
    # Kudos: Adapted from Cirq

    cv_interchangeable = True

    def __init__(self, theta: Variable, phi: Variable, q0: Qubit, q1: Qubit):
        """
        Args:
            theta: Swap angle on the span(|01⟩, |10⟩) subspace, in radians.
            phi: Phase angle, in radians, applied to |11⟩ state.
        """
        super().__init__(params=[theta, phi], qubits=[q0, q1])

    @cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        phi = var.asfloat(self.param("phi"))

        unitary = [
            [1, 0, 0, 0],
            [0, np.cos(theta), -1.0j * np.sin(theta), 0],
            [0, -1.0j * np.sin(theta), np.cos(theta), 0],
            [0, 0, 0, np.exp(-1j * phi)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "FSim":
        return self ** -1

    def __pow__(self, t: Variable) -> "FSim":
        theta, phi = self.params
        return FSim(theta * t, phi * t, *self.qubits)


# end class FSim


# Kudos: Adapted from OpenFermion-Cirq
# https://github.com/quantumlib/OpenFermion-Cirq/blob/master/openfermioncirq/gates/common_gates.py
class FSwap(StdGate):
    r"""Fermionic swap gate. It swaps adjacent fermionic modes in
    the Jordan-Wigner representation.
    Locally equivalent to iSwap and ``Can(1/2, 1/2, 0)``

    .. math::
        \text{FSwap} = \begin{pmatrix}
                        1 & 0 & 0 & 0 \\
                        0 & 0 & i & 0 \\
                        0 & i & 0 & 0 \\
                        0 & 0 & 0 & -1
                       \end{pmatrix}

    """

    cv_interchangable = True
    cv_tensor_structure = "monomial"
    _diagram_labels = [SWAP_TARGET + "ᶠ", SWAP_TARGET + "ᶠ"]

    def __init__(self, q0: Qubit, q1: Qubit):
        super().__init__(qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        return Swap(q0, q1).hamiltonian + CZ(q0, q1).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        return FSwapPow(1, *self.qubits).tensor

    @property
    def H(self) -> "FSwapPow":
        return FSwapPow(-1, *self.qubits)

    def __pow__(self, t: Variable) -> "FSwapPow":
        return FSwapPow(t, *self.qubits)


# End class FSwap


# Kudos: Adapted from OpenFermion-Cirq
# https://github.com/quantumlib/OpenFermion-Cirq/blob/master/openfermioncirq/gates/common_gates.py
class FSwapPow(StdGate):
    """Powers of the fermionic swap (FSwap) gate.
    Locally equivalent to ``Can(t/2, t/2, 0)``
    """

    cv_interchangable = True
    _diagram_labels = [SWAP_TARGET + "ᶠ^{t}", SWAP_TARGET + "ᶠ^{t}"]

    def __init__(self, t: Variable, q0: Qubit, q1: Qubit):
        super().__init__(params=[t], qubits=[q0, q1])

    @property
    def hamiltonian(self) -> Pauli:
        q0, q1 = self.qubits
        return self.param("t") * FSwap(q0, q1).hamiltonian

    @cached_property
    def tensor(self) -> QubitTensor:
        t = self.float_param("t")
        c = np.cos(np.pi * t / 2)
        s = np.sin(np.pi * t / 2)
        g = np.exp(0.5j * np.pi * t)
        p = np.exp(1.0j * np.pi * t)
        unitary = [
            [1, 0, 0, 0],
            [0, g * c, -1.0j * g * s, 0],
            [0, -1.0j * g * s, g * c, 0],
            [0, 0, 0, p],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "FSwapPow":
        return self ** -1

    def __pow__(self, e: Variable) -> "FSwapPow":
        (t,) = self.params
        return FSwapPow(e * t, *self.qubits)


# End class FSwapPow


class Sycamore(StdGate):
    r"""The Sycamore gate is a two-qubit gate equivalent to
    ``FSim(π/2, π/6)``, and locally equivalent to ``Can(1/2, 1/2, 1/6)``.

    This gate was used to demonstrate quantum on Google's Sycamore chip.

    .. math::
        \text{Sycamore}() =
            \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & 0& -i  & 0 \\
                0 & -i  & 0& 0 \\
                0 & 0 & 0 & \exp(- i \pi/6)
            \end{pmatrix}

    Ref:
         https://www.nature.com/articles/s41586-019-1666-5
    """
    # Kudos: Adapted from Cirq

    cv_interchangeable = True
    cv_tensor_structure = "monomial"

    def __init__(self, q0: Qubit, q1: Qubit):
        super().__init__(qubits=[q0, q1])

    @cached_property
    def tensor(self) -> QubitTensor:
        return FSim(np.pi / 2, np.pi / 6, 0, 1).tensor

    @property
    def H(self) -> "FSim":
        return self ** -1

    def __pow__(self, t: Variable) -> "FSim":
        return FSim(t * np.pi / 2, t * np.pi / 6, *self.qubits)


# end class Sycamore


# fin



#!/usr/bin/env python

# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""QuantumFlow: Validate and display various gate translations."""

# Note: Used not only as an illustrative example, but as part of
# the testing suite, and as a source for circuit diagrams in docstrings.

from itertools import zip_longest

import numpy as np

import quantumflow as qf
from quantumflow.translate import translation_source_gate
from quantumflow.visualization import kwarg_to_symbol as syms


def _check_circuit_translations():
    # Concrete values with which to test that circuits are
    # functionally identical.
    concrete = {name: np.random.uniform(-4, 4) for name in syms.values()}

    for name, trans in qf.TRANSLATORS.items():
        gatet = translation_source_gate(trans)
        args = [syms[a] for a in gatet.cv_args]

        gate = gatet(*args, *range(gatet.cv_qubit_nb))
        circ0 = qf.Circuit([gate])
        circ1 = qf.Circuit(trans(gate))

        # FIXME: Fails if no doc string
        annote = trans.__doc__.splitlines()[0]

        _print_circuit_identity(annote, circ0, circ1)

        circ0f = circ0.resolve(concrete)
        circ1f = circ1.resolve(concrete)
        assert qf.gates_close(circ0f.asgate(), circ1f.asgate())


def _print_circuit_identity(
    name, circ0, circ1, min_col_width=0, col_sep=5, left_margin=8
):

    print()
    print("", name)
    print()

    circ0 = qf.Circuit(circ0)
    circ1 = qf.Circuit(circ1)

    gates0 = qf.circuit_to_diagram(circ0, qubit_labels=False).splitlines()
    gates1 = qf.circuit_to_diagram(circ1, qubit_labels=False).splitlines()

    for gate0, gate1 in zip_longest(gates0, gates1, fillvalue=""):
        line = (" " * col_sep).join(
            [gate0.ljust(min_col_width), gate1.ljust(min_col_width)]
        )
        line = (" " * left_margin) + line
        line = line.rstrip()
        print(line)

    print()
    print()


if __name__ == "__main__":
    print()
    print("# Validate and display various gate translations  " "(up to global phase)")
    print()
    _check_circuit_translations()



#!/usr/bin/env python
"""
QuantumFlow: Examples of compiling circuits to native gates
"""

import quantumflow as qf

example_circuits = [
    ["3-qubit addition", qf.addition_circuit([0, 1, 2], [3, 4, 5], [6, 7])],
    ["7-qubit QFT", qf.QFTGate([0, 1, 2, 3, 4, 5, 6, 7]).ascircuit()],  # FIXME
]

for title, example in example_circuits:
    print()
    print(title)
    print(qf.circuit_to_diagram(example))
    print("Gate count:", example.size())

    print()
    print("Simplified circuit")
    circ = qf.compile_circuit(example)
    print(qf.circuit_to_diagram(circ, transpose=True))

    qf.circuit_to_image(circ).show()
    dagc = qf.DAGCircuit(circ)
    print("Gate depth", dagc.depth(local=False))
    print("Operation count", qf.count_operations(dagc))



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import glob
import io
import subprocess

import quantumflow as qf

REPS = 16


def test_version() -> None:
    assert qf.__version__


def test_about() -> None:
    out = io.StringIO()
    qf.about(out)
    print(out)


def test_about_main() -> None:
    rval = subprocess.call(["python", "-m", "quantumflow.about"])
    assert rval == 0


def test_copyright() -> None:
    """Check that source code files contain a copyright line"""
    exclude = set(["quantumflow/version.py"])
    for fname in glob.glob("quantumflow/**/*.py", recursive=True):
        if fname in exclude:
            continue
        print("Checking " + fname + " for copyright header...  ", end="")

        with open(fname) as f:
            for line in f.readlines():
                if not line.strip():
                    continue
                assert line.startswith("# Copyright")
                break
        print("passed")


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""
QuantumFlow: Translate, transform, and compile circuits.
"""

# Note: Beta Prototype

from typing import Callable, Generator, Set, Tuple

from .circuits import Circuit
from .dagcircuit import DAGCircuit
from .info import almost_identity
from .ops import Gate, Operation
from .stdgates import CZ, ZZ, H, XPow, YPow, ZPow
from .translate import (
    circuit_translate,
    translate_ccnot_to_cnot,
    translate_cnot_to_cz,
    translate_cphase_to_zz,
    translate_cswap_to_ccnot,
    translate_hadamard_to_zxz,
    translate_invt_to_tz,
    translate_invv_to_tx,
    translate_t_to_tz,
    translate_tx_to_zxzxz,
    translate_v_to_tx,
    translate_zz_to_cnot,
)


# FIXME: transpile instead of compile?
def compile_circuit(circ: Circuit) -> Circuit:
    """Compile a circuit to standard gate set (CZ, X^0.5, ZPow),
    simplifying circuit where possible.
    """
    # FIXME: Should be automagic translations
    # Convert multi-qubit gates to CZ gates
    trans = [
        translate_cswap_to_ccnot,
        translate_ccnot_to_cnot,
        translate_cphase_to_zz,
        translate_cnot_to_cz,
        translate_t_to_tz,
        translate_invt_to_tz,
        translate_zz_to_cnot,
        translate_v_to_tx,
        translate_invv_to_tx,
    ]
    circ = circuit_translate(circ, trans)

    dagc = DAGCircuit(circ)
    remove_identites(dagc)
    merge_hadamards(dagc)
    convert_HZH(dagc)

    # Standardize 1-qubit gates
    circ = Circuit(dagc)
    circ = circuit_translate(circ, [translate_hadamard_to_zxz])
    circ = circuit_translate(circ, [translate_tx_to_zxzxz], recurse=False)

    # Gather and merge ZPow gates
    dagc = DAGCircuit(circ)
    retrogress_tz(dagc)
    merge_tz(dagc)
    remove_identites(dagc)

    circ = Circuit(dagc)

    return circ


def find_pattern(
    dagc: DAGCircuit,
    gateset1: Set,
    gateset2: Set,
) -> Generator[Tuple[Operation, Operation], None, None]:
    """Find where a gate from gateset1 is followed by a gate from gateset2 in
    a DAGCircuit"""
    for elem2 in dagc:
        if type(elem2) not in gateset2:
            continue

        for q2 in elem2.qubits:
            elem1 = dagc.prev_element(elem2, q2)
            if type(elem1) not in gateset1:
                continue
            yield (elem1, elem2)


def remove_element(dagc: DAGCircuit, elem: Operation) -> None:
    """Remove a node from a DAGCircuit"""

    for qubit in elem.qubits:
        prv = dagc.prev_element(elem, qubit)
        nxt = dagc.next_element(elem, qubit)
        dagc.graph.add_edge(prv, nxt, key=qubit)
    dagc.graph.remove_node(elem)


def remove_identites(dagc: DAGCircuit) -> None:
    """Remove identities from a DAGCircuit"""
    for elem in dagc:
        if isinstance(elem, Gate) and almost_identity(elem):
            remove_element(dagc, elem)


def merge_hadamards(dagc: DAGCircuit) -> None:
    """Merge and remove neighboring Hadamard gates"""
    for elem1, elem2 in find_pattern(dagc, {H}, {H}):
        remove_element(dagc, elem1)
        remove_element(dagc, elem2)


def merge_tx(dagc: DAGCircuit) -> None:
    """Merge neighboring ZPow gates"""
    _merge_turns(dagc, XPow)


def merge_ty(dagc: DAGCircuit) -> None:
    """Merge neighboring ZPow gates"""
    _merge_turns(dagc, YPow)


def merge_tz(dagc: DAGCircuit) -> None:
    """Merge neighboring ZPow gates"""
    _merge_turns(dagc, ZPow)


def _merge_turns(dagc: DAGCircuit, gate_class: Callable) -> None:
    for gate0, gate1 in find_pattern(dagc, {gate_class}, {gate_class}):
        t = gate0.param("t") + gate1.param("t")
        (qubit,) = gate0.qubits
        gate = gate_class(t, qubit)

        prv = dagc.prev_element(gate0)
        nxt = dagc.next_element(gate1)
        dagc.graph.add_edge(prv, gate, key=qubit)
        dagc.graph.add_edge(gate, nxt, key=qubit)

        dagc.graph.remove_node(gate0)
        dagc.graph.remove_node(gate1)


def retrogress_tz(dagc: DAGCircuit) -> None:
    """Commute ZPow gates as far backward in the circuit as possible"""
    G = dagc.graph
    again = True
    while again:
        again = False
        for elem1, elem2 in find_pattern(dagc, {ZZ, CZ}, {ZPow}):
            (q,) = elem2.qubits
            elem0 = dagc.prev_element(elem1, q)
            elem3 = dagc.next_element(elem2, q)

            G.remove_edge(elem0, elem1, q)
            G.remove_edge(elem1, elem2, q)
            G.remove_edge(elem2, elem3, q)

            G.add_edge(elem0, elem2, key=q)
            G.add_edge(elem2, elem1, key=q)
            G.add_edge(elem1, elem3, key=q)
            again = True


# TODO: Rename? merge_hzh
# TODO: larger pattern, simplifying sequences of 1-qubit Clifford gates
def convert_HZH(dagc: DAGCircuit) -> None:
    """Convert a sequence of H-ZPow-H gates to a XPow gate"""
    for elem2, elem3 in find_pattern(dagc, {ZPow}, {H}):
        elem1 = dagc.prev_element(elem2)
        if not isinstance(elem1, H):
            continue

        prv = dagc.prev_element(elem1)
        nxt = dagc.next_element(elem3)

        t = elem2.param("t")
        (q0,) = elem2.qubits
        gate = XPow(t, q0)

        dagc.graph.remove_node(elem1)
        dagc.graph.remove_node(elem2)
        dagc.graph.remove_node(elem3)

        dagc.graph.add_edge(prv, gate, key=q0)
        dagc.graph.add_edge(gate, nxt, key=q0)


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
========================
Miscellaneous Operations
========================


.. currentmodule:: quantumflow

Various standard operations on quantum states, which aren't gates,
channels, circuits, or DAGCircuit's.

.. autoclass :: Moment

.. autoclass :: Measure
.. autoclass :: Reset
.. autoclass :: Initialize
.. autoclass :: Barrier
.. autoclass :: Store
.. autoclass :: If
.. autoclass :: Display
.. autoclass :: StateDisplay
.. autoclass :: ProbabilityDisplay
.. autoclass :: DensityDisplay
.. autoclass :: Projection

"""

import textwrap
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    Sequence,
    Tuple,
    Union,
)

import numpy as np

from . import tensors, utils
from .circuits import Circuit
from .config import CIRCUIT_INDENT
from .gates import P0, P1
from .ops import Channel, Gate, Operation, Unitary
from .qubits import Qubit, Qubits
from .states import Density, State
from .tensors import QubitTensor
from .var import Variable

__all__ = [
    "Moment",
    "Measure",
    "Reset",
    "Initialize",
    "Barrier",
    "Store",
    "If",
    "Display",
    "StateDisplay",
    "ProbabilityDisplay",
    "DensityDisplay",
    "Projection",
]


class Moment(Sequence, Operation):
    """
    Represents a collection of Operations that operate on disjoint qubits,
    so that they may be applied at the same moment of time.
    """

    def __init__(self, *elements: Union[Iterable[Operation], Operation]) -> None:
        if len(elements) == 1 and isinstance(elements[0], Iterable):
            elements = elements[0]  # type: ignore

        circ = Circuit(Circuit(elements).flat())  # type: ignore
        qbs = list(q for elem in circ for q in elem.qubits)
        if len(qbs) != len(set(qbs)):
            raise ValueError("Qubits of operations within Moments " "must be disjoint.")

        super().__init__(qubits=qbs)
        self._circ = circ

    def __getitem__(self, key: Union[int, slice]) -> Any:
        return self._circ[key]

    def __len__(self) -> int:
        return self._circ.__len__()

    def __iter__(self) -> Iterator[Operation]:
        yield from self._circ

    def run(self, ket: State = None) -> State:
        return self._circ.run(ket)

    def evolve(self, rho: Density = None) -> Density:
        return self._circ.evolve(rho)

    def asgate(self) -> "Gate":
        return self._circ.asgate()

    def aschannel(self) -> "Channel":
        return self._circ.aschannel()

    @property
    def H(self) -> "Moment":
        return Moment(self._circ.H)

    def __str__(self) -> str:
        circ_str = "\n".join([str(elem) for elem in self])
        circ_str = textwrap.indent(circ_str, " " * CIRCUIT_INDENT)
        return "\n".join([self.name, circ_str])

    def on(self, *qubits: Qubit) -> "Moment":
        return Moment(Circuit(self).on(*qubits))

    def rewire(self, labels: Dict[Qubit, Qubit]) -> "Moment":
        return Moment(Circuit(self).rewire(labels))

    @property
    def params(self) -> Tuple[Variable, ...]:
        return tuple(item for elem in self for item in elem.params)

    def param(self, name: str) -> Variable:
        raise ValueError("Cannot lookup parameters by name for composite operations")


# end class Moment


class Measure(Operation):
    """Measure a quantum bit and copy result to a classical bit"""

    _diagram_labels = ["M({cbit})"]

    def __init__(self, qubit: Qubit, cbit: Hashable = None) -> None:
        if cbit is None:
            cbit = qubit

        super().__init__(qubits=[qubit])
        self.qubit = qubit
        self.cbit = cbit

    def __str__(self) -> str:
        if self.cbit != self.qubit:
            return f"{self.name} {self.qubit} {self.cbit}"
        return f"{self.name} {self.qubit}"

    def run(self, ket: State) -> State:
        prob_zero = P0(self.qubit).run(ket).norm()

        # generate random number to 'roll' for measurement
        if np.random.random() < prob_zero:
            ket = P0(self.qubit).run(ket).normalize()
            ket = ket.store({self.cbit: 0})
        else:  # measure one
            ket = P1(self.qubit).run(ket).normalize()
            ket = ket.store({self.cbit: 1})
        return ket

    def evolve(self, rho: Density) -> Density:
        p0 = P0(self.qubit).aschannel()
        p1 = P1(self.qubit).aschannel()

        prob_zero = p0.evolve(rho).norm()

        # generate random number to 'roll' for measurement
        if np.random.random() < prob_zero:
            rho = p0.evolve(rho).normalize()
            rho = rho.store({self.cbit: 0})
        else:  # measure one
            rho = p1.evolve(rho).normalize()
            rho = rho.store({self.cbit: 1})
        return rho


# FIXME: Can't have zero qubits
# Having no qubits specified screws up visualization
# and dagc
class Reset(Operation):
    r"""An operation that resets qubits to zero irrespective of the
    initial state.
    """

    _diagram_labels = ["┤ ⟨0|"]
    _diagram_noline = True

    def __init__(self, *qubits: Qubit) -> None:
        if not qubits:
            qubits = ()
        super().__init__(qubits)

        self._gate = Unitary(tensor=[[1, 1], [0, 0]], qubits=[0])

    def run(self, ket: State) -> State:
        if self.qubits:
            qubits = self.qubits
        else:
            qubits = ket.qubits

        for q in qubits:
            gate = self._gate.on(q)
            ket = gate.run(ket)
        ket = ket.normalize()
        return ket

    def evolve(self, rho: Density) -> Density:
        # TODO
        raise TypeError("Not yet implemented")

    def asgate(self) -> Gate:
        raise TypeError("Reset not convertible to Gate")

    # FIXME?
    def aschannel(self) -> Channel:
        raise TypeError("Reset not convertible to Channel")

    def __str__(self) -> str:
        if self.qubits:
            return "Reset " + " ".join([str(q) for q in self.qubits])
        return "Reset"


class Initialize(Operation):
    """ An operation that initializes the quantum state"""

    def __init__(self, ket: State):
        self._ket = ket
        self._qubits = ket.qubits  # FIXME
        super().__init__(ket.qubits)

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return self._ket.tensor

    def run(self, ket: State) -> State:
        return self._ket.permute(ket.qubits)

    def evolve(self, rho: Density) -> Density:
        return self._ket.permute(rho.qubits).asdensity()

    # TODO: aschannel? __str___?


# TODO: Could be a Gate
# FIXME: Interface
class Barrier(Operation):
    """An operation that prevents reordering of operations across the barrier.
    Has no effect on the quantum state."""

    interchangable = True
    _diagram_labels = ["┼"]
    _diagram_noline = True

    def __init__(self, *qubits: Qubit) -> None:
        super().__init__(qubits=qubits)

    @property
    def H(self) -> "Barrier":
        return self  # Hermitian

    def run(self, ket: State) -> State:
        return ket  # NOP

    def evolve(self, rho: Density) -> Density:
        return rho  # NOP

    def __str__(self) -> str:
        return self.name + " " + " ".join(str(q) for q in self.qubits)


# FIXME: Does not work as written?
class Projection(Operation):
    """A projection operator, represented as a sequence of state vectors"""

    # TODO: evolve(), asgate(), aschannel()

    def __init__(self, states: Sequence[State]):
        self.states = states

        qbs = [q for state in self.states for q in state.qubits]  # gather
        qbs = list(set(qbs))  # unique
        qbs = sorted(qbs)  # sort

        super().__init__(qbs)

    def run(self, ket: State) -> State:
        """Apply the action of this operation upon a pure state"""
        tensor = sum(
            state.tensor * tensors.inner(state.tensor, ket.tensor)
            for state in self.states
        )
        return State(tensor, qubits=ket.qubits)

    @property
    def H(self) -> "Projection":
        return self  # pragma: no cover  # TESTME


# end class Projection


class Store(Operation):
    """Store a value in the classical memory of the state."""

    def __init__(self, key: Hashable, value: Any, qubits: Qubits = ()) -> None:
        super().__init__(qubits=qubits)
        self.key = key
        self.value = value

    def run(self, ket: State) -> State:
        return ket.store({self.key: self.value})

    def evolve(self, rho: Density) -> Density:
        return rho.store({self.key: self.value})


# end class Store


class If(Operation):
    """
    Look up key in classical memory, and apply the given
    quantum operation only if the truth value matches.
    """

    def __init__(self, elem: Operation, key: Hashable, value: bool = True) -> None:
        super().__init__(qubits=elem.qubits)
        self.element = elem
        self.key = key
        self.value = value

    def run(self, ket: State) -> State:
        if ket.memory[self.key] == self.value:
            ket = self.element.run(ket)
        return ket

    def evolve(self, rho: Density) -> Density:
        if rho.memory[self.key] == self.value:
            rho = self.element.evolve(rho)
        return rho


# end class If


class Display(Operation):
    """A Display is an operation that extracts information from the
    quantum state and stores it in classical memory, without performing
    any effect on the qubits.
    """

    # Terminology 'Display' used by Quirk (https://algassert.com/quirk)
    # and cirq (cirq/ops/display.py).
    def __init__(self, key: Hashable, action: Callable, qubits: Qubits = ()) -> None:
        super().__init__(qubits=qubits)
        self.key = key
        self.action = action

    def run(self, ket: State) -> State:
        return ket.store({self.key: self.action(ket)})

    def evolve(self, rho: Density) -> Density:
        return rho.store({self.key: self.action(rho)})


# end class Display


class StateDisplay(Display):
    """
    Store a copy of the state in the classical memory. (This operation
    can be memory intensive, since it stores the entire quantum state.)
    """

    def __init__(self, key: Hashable, qubits: Qubits = ()) -> None:
        super().__init__(key, lambda x: x, qubits=qubits)


# TODO: Act on qubit subspace
class ProbabilityDisplay(Display):
    """
    Store the state probabilities in classical memory.
    """

    def __init__(self, key: Hashable, qubits: Qubits = ()) -> None:
        super().__init__(key, lambda state: state.probabilities(), qubits=qubits)


# TESTME
class DensityDisplay(Display):
    """
    Store the density matrix of given qubits in classical memory.
    """

    def __init__(self, key: Hashable, qubits: Qubits) -> None:
        super().__init__(key, lambda state: state.asdensity(qubits), qubits=qubits)


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


import networkx as nx
import numpy as np
import pytest
import scipy.linalg

import quantumflow as qf
from quantumflow import var


def test_identitygate() -> None:
    qubits = [3, 4, 5, 6, 7, 8]
    gate = qf.IdentityGate(qubits)
    ket0 = qf.random_state(qubits)
    ket1 = gate.run(ket0)
    assert qf.states_close(ket0, ket1)

    circ = qf.Circuit(gate.decompose())
    assert len(circ) == 6

    for n in range(1, 6):
        assert qf.IdentityGate(list(range(n))).qubit_nb == n

    assert gate.hamiltonian.is_zero()

    assert gate ** 2 is gate


def test_control_gate() -> None:
    gate0 = qf.ControlGate([0], qf.X(1))
    gate1 = qf.CNot(0, 1)
    assert qf.gates_close(gate0, gate1)

    gateb = qf.ControlGate([1], qf.X(0))
    gate2 = qf.CNot(1, 0)
    assert qf.gates_close(gateb, gate2)

    gate3 = qf.ControlGate([0], qf.Y(1))
    gate4 = qf.CY(0, 1)
    assert qf.gates_close(gate3, gate4)

    gate5 = qf.ControlGate([0], qf.Z(1))
    gate6 = qf.CZ(0, 1)
    assert qf.gates_close(gate5, gate6)

    gate7 = qf.ControlGate([0], qf.H(1))
    gate8 = qf.CH(0, 1)
    assert qf.gates_close(gate7, gate8)

    gate9 = qf.ControlGate([0, 1], qf.X(2))
    gate10 = qf.CCNot(0, 1, 2)
    assert qf.gates_close(gate9, gate10)

    gate11 = qf.ControlGate([0], qf.Swap(1, 2))
    gate12 = qf.CSwap(0, 1, 2)
    assert qf.gates_close(gate11, gate12)


def test_qftgate() -> None:
    circ = qf.Circuit()
    circ += qf.X(2)
    circ += qf.QFTGate([0, 1, 2])

    ket = qf.zero_state(3)
    ket = circ.run(ket)

    true_qft = qf.State(
        [
            0.35355339 + 0.0j,
            0.25000000 + 0.25j,
            0.00000000 + 0.35355339j,
            -0.25000000 + 0.25j,
            -0.35355339 + 0.0j,
            -0.25000000 - 0.25j,
            0.00000000 - 0.35355339j,
            0.25000000 - 0.25j,
        ]
    )

    assert qf.states_close(ket, true_qft)

    assert isinstance(qf.QFTGate([0, 1, 2]).H, qf.InvQFTGate)
    assert isinstance(qf.QFTGate([0, 1, 2]).H.H, qf.QFTGate)

    qf.QFTGate([0, 1, 2]).H.tensor


def test_multiswapgate() -> None:
    # Should be same as a swap.
    perm0 = qf.MultiSwapGate([0, 1], [1, 0])
    gate0 = qf.Swap(0, 1)
    assert qf.gates_close(perm0.asgate(), gate0)
    assert qf.gates_close(perm0.asgate(), perm0.H.asgate())

    perm1 = qf.MultiSwapGate.from_gates(qf.Circuit([gate0]))
    assert qf.gates_close(perm0.asgate(), perm1.asgate())

    perm2 = qf.MultiSwapGate.from_gates(qf.Circuit([perm1]))
    assert qf.gates_close(perm0, perm2)

    with pytest.raises(ValueError):
        qf.MultiSwapGate.from_gates(qf.Circuit(qf.CNot(0, 1)))

    N = 8
    qubits_in = list(range(N))
    qubits_out = np.random.permutation(qubits_in)

    permN = qf.MultiSwapGate(qubits_in, qubits_out)
    assert qf.gates_close(perm0.asgate(), perm1.asgate())
    iden = qf.Circuit([permN, permN.H])
    assert qf.almost_identity(iden.asgate())
    assert qf.circuits_close(iden, qf.Circuit([qf.IdentityGate(qubits_in)]))

    swaps = qf.Circuit(permN.decompose())
    # Add identity so we don't lose qubits
    swaps += qf.IdentityGate(permN.qubits_in)
    permN2 = qf.MultiSwapGate.from_gates(swaps)

    assert qf.circuits_close(swaps, qf.Circuit([permN]))
    assert qf.circuits_close(swaps, qf.Circuit([permN2]))
    assert qf.circuits_close(qf.Circuit([permN]), qf.Circuit([permN2]))

    with pytest.raises(ValueError):
        _ = qf.MultiSwapGate([0, 1], [1, 2])

    # Channels
    assert qf.channels_close(perm0.aschannel(), gate0.aschannel())

    rho0 = qf.random_state([0, 1, 3]).asdensity()
    rho1 = perm0.evolve(rho0)
    rho2 = gate0.aschannel().evolve(rho0)
    assert qf.densities_close(rho1, rho2)


def test_reversequbits() -> None:
    rev = qf.ReversalGate([0, 1, 2, 3, 4])
    perm = qf.MultiSwapGate([0, 1, 2, 3, 4], [4, 3, 2, 1, 0])
    assert qf.circuits_close(qf.Circuit(rev.decompose()), qf.Circuit(perm.decompose()))


def test_rotatequbits() -> None:
    rev = qf.CircularShiftGate([0, 1, 2, 3, 4], 2)
    perm = qf.MultiSwapGate([0, 1, 2, 3, 4], [2, 3, 4, 0, 1])
    assert qf.circuits_close(qf.Circuit(rev.decompose()), qf.Circuit(perm.decompose()))


def test_PauliGate() -> None:
    pauli0 = 0.5 * np.pi * qf.sX(0) * qf.sX(1)

    alpha = 0.4
    circ = qf.PauliGate(pauli0, alpha)
    coords = qf.canonical_coords(circ.asgate())
    assert np.isclose(coords[0], 0.4)

    pauli1 = np.pi * qf.sX(0) * qf.sX(1) * qf.sY(2) * qf.sZ(3)
    _ = qf.PauliGate(pauli1, alpha)

    top2 = nx.star_graph(4)
    pauli2 = 0.5 * np.pi * qf.sX(1) * qf.sY(2) * qf.sZ(3)
    _ = qf.PauliGate(pauli2, alpha).decompose(top2)

    alpha = 0.2
    top3 = nx.star_graph(4)
    pauli3 = 0.5 * np.pi * qf.sX(1) * qf.sX(2)
    circ3 = qf.Circuit(qf.PauliGate(pauli3, alpha).decompose(top3))

    assert qf.circuits_close(circ3, qf.Circuit([qf.I(0), qf.XX(alpha, 1, 2)]))

    qf.PauliGate(qf.sI(0), alpha).decompose(top2)

    with pytest.raises(ValueError):
        pauli4 = 0.5j * np.pi * qf.sX(1) * qf.sX(2)
        _ = qf.Circuit(qf.PauliGate(pauli4, alpha).decompose(top3))

    top4 = nx.DiGraph()
    nx.add_path(top4, [3, 2, 1, 0])
    circ3 = qf.Circuit(qf.PauliGate(pauli3, alpha).decompose(top4))


def test_PauliGate_more() -> None:

    alphas = [0.1, 2.0, -3.14, -0.4]
    paulis = [
        qf.sZ(0) + 1,
        qf.sY(0),
        qf.sX(0),
        0.5 * np.pi * qf.sZ(0) * qf.sZ(1),
        0.5 * np.pi * qf.sX(0) * qf.sZ(1),
    ]

    for alpha in alphas:
        for pauli in paulis:
            circ = qf.PauliGate(pauli, alpha)
            qbs = circ.qubits
            str(pauli)
            op = pauli.asoperator(qbs)
            U = scipy.linalg.expm(-1.0j * alpha * op)
            gate = qf.Unitary(U, qbs)
            assert qf.gates_close(gate, circ.asgate())

    pauli = qf.sX(0) + qf.sZ(0)
    with pytest.raises(ValueError):
        qf.Circuit(qf.PauliGate(pauli, 0.2).decompose())


def test_PauliGate_resolve() -> None:
    alpha = var.Symbol("alpha")
    g = qf.PauliGate(qf.sZ(0), alpha)
    r = g.resolve(subs={"alpha": 0.3})
    assert r.alpha == 0.3

    g = qf.PauliGate(qf.sZ(0), 0.3)
    assert g.resolve(subs={"alpha": 0.3}) is g


def test_PauliGate_pow() -> None:
    alpha = 0.4
    gate0 = qf.PauliGate(qf.sZ(0), alpha)
    gate1 = gate0 ** 0.3
    gate2 = qf.Unitary(gate0.tensor, gate0.qubits) ** 0.3
    assert qf.gates_close(gate1, gate2)
    assert qf.gates_close(gate1.H, gate2 ** -1)

    gate3 = qf.unitary_from_hamiltonian(gate0.hamiltonian, qubits=gate0.qubits)
    assert qf.gates_close(gate0, gate3)

    s = str(gate0)
    print(s)


# fin



# Copyright 2020-, Gavin E. Crooks and contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


"""
====================
Information Measures
====================

Measures on vectors, states, and quantum operations.

.. contents:: :local:
.. currentmodule:: quantumflow


Measures in Hilbert Space
##########################

.. autofunction:: fubini_study_angle
.. autofunction:: fubini_study_fidelity
.. autofunction:: fubini_study_close

State Measures
###############

.. autofunction:: state_angle
.. autofunction:: states_close
.. autofunction:: state_fidelity


Mixed state Measures
######################

.. autofunction:: density_angle
.. autofunction:: purity
.. autofunction:: densities_close
.. autofunction:: fidelity
.. autofunction:: bures_distance
.. autofunction:: bures_angle
.. autofunction:: entropy
.. autofunction:: mutual_info


Gate Measures
##############

.. autofunction:: gate_angle
.. autofunction:: gates_close
.. autofunction:: gates_phase_close
.. autofunction:: gates_commute
.. autofunction:: almost_unitary
.. autofunction:: almost_hermitian
.. autofunction:: almost_identity


Channel Measures
#################

.. autofunction:: channel_angle
.. autofunction:: channels_close
.. autofunction:: average_gate_fidelity
.. autofunction:: almost_unital
"""

import numpy as np
import scipy.stats
from scipy.linalg import sqrtm  # matrix square root

from . import tensors
from .channels import Kraus
from .circuits import Circuit
from .config import ATOL
from .modules import IdentityGate
from .ops import Channel, Gate
from .qubits import Qubits
from .states import Density, State, random_state
from .tensors import QubitTensor

__all__ = (
    "fubini_study_angle",
    "fubini_study_fidelity",
    "fubini_study_close",
    "state_fidelity",
    "state_angle",
    "states_close",
    "purity",
    "fidelity",
    "bures_distance",
    "bures_angle",
    "density_angle",
    "densities_close",
    "entropy",
    "mutual_info",
    "gate_angle",
    "channel_angle",
    "gates_close",
    "gates_phase_close",
    "gates_commute",
    "channels_close",
    "circuits_close",
    "trace_distance",
    "average_gate_fidelity",
    "almost_unitary",
    "almost_identity",
    "almost_hermitian",
    "almost_unital",
)


# -- Measures on Hilbert space ---


def fubini_study_angle(vec0: QubitTensor, vec1: QubitTensor) -> QubitTensor:
    """Calculate the Fubini–Study metric between elements of a Hilbert space.

    The Fubini–Study metric is a distance measure between vectors in a
    projective Hilbert space. For gates this space is the Hilbert space of
    operators induced by the Hilbert-Schmidt inner product.
    For 1-qubit rotation gates, Rx, Ry and Rz, this is half the angle (theta)
    in the Bloch sphere.

    The Fubini–Study metric between states is equal to the Burr angle
    between pure states.
    """
    fs_fidelity = fubini_study_fidelity(vec0, vec1)
    return np.arccos(fs_fidelity)


def fubini_study_fidelity(vec0: QubitTensor, vec1: QubitTensor) -> QubitTensor:
    """
    Cosine of the Fubini–Study metric.
    """
    # Suffers from less floating point errors compared to fubini_study_angle

    hs01 = tensors.inner(vec0, vec1)  # Hilbert-Schmidt inner product
    hs00 = tensors.inner(vec0, vec0)
    hs11 = tensors.inner(vec1, vec1)
    ratio = np.absolute(hs01) / np.sqrt(np.absolute(hs00 * hs11))
    fid = np.minimum(ratio, 1.0)  # Compensate for rounding errors.
    return fid


def fubini_study_close(
    vec0: QubitTensor, vec1: QubitTensor, atol: float = ATOL
) -> bool:
    """Return True if vectors are close in the projective Hilbert space.

    Similarity is measured with the Fubini–Study metric.
    """

    return 1 - fubini_study_fidelity(vec0, vec1) <= atol


# -- Measures on pure states ---


def state_fidelity(ket0: State, ket1: State) -> QubitTensor:
    """Return the quantum fidelity between pure states."""
    ket1 = ket1.permute(ket0.qubits)
    tensor = np.absolute(tensors.inner(ket0.tensor, ket1.tensor)) ** 2
    return tensor


def state_angle(ket0: State, ket1: State) -> QubitTensor:
    """The Fubini-Study angle between states.

    Equal to the Burrs angle for pure states.
    """
    ket1 = ket1.permute(ket0.qubits)
    return fubini_study_angle(ket0.tensor, ket1.tensor)


def states_close(ket0: State, ket1: State, atol: float = ATOL) -> bool:
    """Returns True if states are almost identical.

    Closeness is measured with the metric Fubini-Study angle.
    """
    ket1 = ket1.permute(ket0.qubits)

    return fubini_study_close(ket0.tensor, ket1.tensor, atol)


# -- Measures on density matrices ---


def purity(rho: Density) -> QubitTensor:
    """
    Calculate the purity of a mixed quantum state.

    Purity, defined as tr(rho^2), has an upper bound of 1 for a pure state,
    and a lower bound of 1/D (where D is the Hilbert space dimension) for a
    competently mixed state.

    Two closely related.info are the linear entropy, 1- purity, and the
    participation ratio, 1/purity.
    """
    tensor = rho.tensor
    N = rho.qubit_nb
    matrix = np.reshape(tensor, [2 ** N, 2 ** N])
    return np.trace(matrix @ matrix)


def fidelity(rho0: Density, rho1: Density) -> float:
    """Return the fidelity F(rho0, rho1) between two mixed quantum states."""
    rho1 = rho1.permute(rho0.qubits)
    op0 = rho0.asoperator()
    op1 = rho1.asoperator()

    fid = np.real((np.trace(sqrtm(sqrtm(op0) @ op1 @ sqrtm(op0)))) ** 2)
    fid = min(fid, 1.0)
    fid = max(fid, 0.0)  # Correct for rounding errors

    return fid


def bures_distance(rho0: Density, rho1: Density) -> float:
    """Return the Bures distance between mixed quantum states"""
    fid = fidelity(rho0, rho1)
    tr0 = np.trace(rho0.asoperator())
    tr1 = np.trace(rho1.asoperator())

    return np.sqrt(tr0 + tr1 - 2.0 * np.sqrt(fid))


def bures_angle(rho0: Density, rho1: Density) -> float:
    """Return the Bures angle between mixed quantum states"""
    return np.arccos(np.sqrt(fidelity(rho0, rho1)))


def density_angle(rho0: Density, rho1: Density) -> QubitTensor:
    """The Fubini-Study angle between density matrices"""
    rho1 = rho1.permute(rho0.qubits)
    return fubini_study_angle(rho0.tensor, rho1.tensor)


def densities_close(rho0: Density, rho1: Density, atol: float = ATOL) -> bool:
    """Returns True if densities are almost identical.

    Closeness is measured with the Fubini-Study fidelity.
    """
    rho1 = rho1.permute(rho0.qubits)
    return fubini_study_close(rho0.tensor, rho1.tensor, atol)


def entropy(rho: Density, base: float = None) -> float:
    """
    Returns the von-Neumann entropy of a mixed quantum state.

    Args:
        rho:    A density matrix
        base:   Optional logarithm base. Default is base e, and entropy is
               .info in nats. For bits set base to 2.

    Returns:
        The von-Neumann entropy of rho
    """
    op = rho.asoperator()
    probs = np.linalg.eigvalsh(op)
    probs = np.maximum(probs, 0.0)  # Compensate for floating point errors
    return scipy.stats.entropy(probs, base=base)


# TESTME
def mutual_info(
    rho: Density, qubits0: Qubits, qubits1: Qubits = None, base: float = None
) -> float:
    """Compute the bipartite von-Neumann mutual information of a mixed
    quantum state.

    Args:
        rho:    A density matrix of the complete system
        qubits0: Qubits of system 0
        qubits1: Qubits of system 1. If none, taken to be all remaining qubits
        base:   Optional logarithm base. Default is base e

    Returns:
        The bipartite von-Neumann mutual information.
    """
    if qubits1 is None:
        qubits1 = tuple(set(rho.qubits) - set(qubits0))

    rho0 = rho.asdensity(qubits0)
    rho1 = rho.asdensity(qubits1)

    ent = entropy(rho, base)
    ent0 = entropy(rho0, base)
    ent1 = entropy(rho1, base)

    return ent0 + ent1 - ent


# TESTME
def trace_distance(rho0: Density, rho1: Density) -> float:
    """
    Compute the trace distance between two mixed states
    :math:`T(rho_0,rho_1) = (1/2)||rho_0-rho_1||_1`
    """
    op0 = rho0.asoperator()
    op1 = rho1.asoperator()
    return 0.5 * np.linalg.norm(op0 - op1, 1)


# Measures on gates


def gate_angle(gate0: Gate, gate1: Gate) -> QubitTensor:
    """The Fubini-Study angle between gates"""
    gate1 = gate1.permute(gate0.qubits)
    return fubini_study_angle(gate0.tensor, gate1.tensor)


def gates_close(gate0: Gate, gate1: Gate, atol: float = ATOL) -> bool:
    """Returns: True if gates are almost identical, up to
    a phase factor.

    Closeness is measured with the Fubini-Study metric.
    """
    gate1 = gate1.permute(gate0.qubits)
    return fubini_study_close(gate0.tensor, gate1.tensor, atol)


# TESTME
def gates_phase_close(gate0: Gate, gate1: Gate, atol: float = ATOL) -> bool:
    """Returns: True if gates are almost identical and
    have almost the same phase.
    """
    gate1 = gate1.permute(gate0.qubits)
    if not gates_close(gate0, gate1):
        return False
    N = gate0.qubit_nb
    phase = np.trace((gate1 @ gate0.H).asoperator()) / 2 ** N
    return np.isclose(phase, 1.0)


# TESTME
def gates_commute(gate0: Gate, gate1: Gate, atol: float = ATOL) -> bool:
    """Returns: True if gates (almost) commute."""
    gate01 = Circuit([gate0, gate1]).asgate()
    gate10 = Circuit([gate1, gate0]).asgate()
    return gates_close(gate01, gate10, atol)


def almost_unitary(gate: Gate) -> bool:
    """Return true if gate is (almost) unitary"""
    return gates_close(gate @ gate.H, IdentityGate(gate.qubits))


def almost_identity(gate: Gate) -> bool:
    """Return true if gate tensor is (almost) the identity"""
    return gates_close(gate, IdentityGate(gate.qubits))


def almost_hermitian(gate: Gate) -> bool:
    """Return true if gate tensor is (almost) Hermitian"""
    return gates_close(gate, gate.H)


# Measures on circuits

# DOCME
def circuits_close(
    circ0: Circuit, circ1: Circuit, atol: float = ATOL, reps: int = 16
) -> bool:
    """Returns: True if circuits are (probably) almost identical.

    We check closeness by running multiple random initial states
    through both circuits and checking that the resultant states are close.
    """
    qubits = circ0.qubits
    if qubits != circ1.qubits:
        return False

    for _ in range(reps):
        ket = random_state(qubits)
        if not states_close(circ0.run(ket), circ1.run(ket), atol):
            return False
    return True


# Measures on channels


def channel_angle(chan0: Channel, chan1: Channel) -> QubitTensor:
    """The Fubini-Study angle between channels"""
    chan1 = chan1.permute(chan0.qubits)
    return fubini_study_angle(chan0.tensor, chan1.tensor)


def channels_close(chan0: Channel, chan1: Channel, atol: float = ATOL) -> bool:
    """Returns: True if channels are almost identical.

    Closeness is measured with the channel angle.
    """
    chan1 = chan1.permute(chan0.qubits)
    return fubini_study_close(chan0.tensor, chan1.tensor, atol)


# TESTME  multiqubits
def average_gate_fidelity(kraus: Kraus, target: Gate = None) -> QubitTensor:
    """Return the average gate fidelity between a noisy gate (specified by a
    Kraus representation of a superoperator), and a purely unitary target gate.

    If the target gate is not specified, default to identity gate.
    """

    if target is None:
        target = IdentityGate(kraus.qubits)
    else:
        if kraus.qubits != target.qubits:
            raise ValueError("Qubits must be same")  # pragma: no cover  # TESTME

    N = kraus.qubit_nb
    d = 2 ** N

    U = target.H.asoperator()

    summand = 0
    for w, K in zip(kraus.weights, kraus.operators):
        summand += np.absolute(np.trace(w * U @ K.asoperator())) ** 2

    return (d + summand) / (d + d ** 2)


# Author: GEC (2019)
def almost_unital(chan: Channel) -> bool:
    """Return true if the channel is (almost) unital."""
    # Unital channels leave the identity unchanged.
    dim = 2 ** chan.qubit_nb
    eye0 = np.eye(dim, dim)
    rho0 = Density(eye0, chan.qubits)
    rho1 = chan.evolve(rho0)
    eye1 = rho1.asoperator()
    return np.allclose(eye0, eye1)


# fin



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.


import pytest

import quantumflow as qf


def test_moment() -> None:
    circ = qf.Circuit()
    circ += qf.X(0)
    circ += qf.Swap(1, 2)

    moment = qf.Moment(circ)

    assert moment.qubits == (0, 1, 2)
    assert moment.run()
    assert moment.evolve()
    assert isinstance(moment.H, qf.Moment)

    circ += qf.Y(0)
    with pytest.raises(ValueError):
        moment = qf.Moment(circ)

    assert moment.asgate()
    assert moment.aschannel()

    circ1 = qf.Circuit(moment)
    assert len(circ1) == 2

    assert isinstance(moment[1], qf.Swap)

    moment1 = moment.on("a", "b", "c")

    moment2 = moment1.rewire({"a": 0, "b": 1, "c": 2})
    assert str(moment) == str(moment2)


def test_moment_params() -> None:
    circ = qf.Circuit()
    circ += qf.X(0) ** 0.3
    circ += qf.Swap(1, 2)
    moment = qf.Moment(circ)

    assert len(moment.params) == 1
    assert moment.params == (0.3,)

    with pytest.raises(ValueError):
        _ = moment.param("theta")


def test_measure() -> None:
    prog = qf.Circuit()
    prog += qf.Measure(0, ("c", 0))
    prog += qf.X(0)
    prog += qf.Measure(0, ("c", 1))
    ket = prog.run()

    assert ket.qubits == (0,)
    assert ket.memory[("c", 0)] == 0
    assert ket.memory[("c", 1)] == 1

    assert str(qf.Measure(4)) == "Measure 4"
    assert str(qf.Measure(4, "c0")) == "Measure 4 c0"


def test_barrier() -> None:
    circ = qf.Circuit()
    circ += qf.Barrier(0, 1, 2)
    circ += qf.Barrier(0, 1, 2).H
    circ.run()
    circ.evolve()

    assert str(qf.Barrier(0, 1, 2)) == "Barrier 0 1 2"


def test_if() -> None:
    circ = qf.Circuit()
    c = ["c0", "c1"]
    circ += qf.Store(c[0], 0)
    circ += qf.Store(c[1], 0)
    circ += qf.If(qf.X(0), c[1], value=False)
    circ += qf.Measure(0, c[0])
    ket = circ.run()
    assert ket.memory[c[0]] == 1
    assert circ.evolve().memory[c[0]] == 1

    circ = qf.Circuit()
    circ += qf.Store(c[0], 0)
    circ += qf.Store(c[1], 0)
    circ += qf.If(qf.X(0), c[1])
    circ += qf.Measure(0, c[0])
    ket = circ.run()
    assert ket.memory[c[0]] == 0
    assert circ.evolve().memory[c[0]] == 0

    circ = qf.Circuit()
    circ += qf.Store(c[0], 0)
    circ += qf.Store(c[1], 0)
    circ += qf.If(qf.X(0), c[1], value=False)
    circ += qf.Measure(0, c[0])
    ket = circ.run()
    assert ket.memory[c[0]] == 1
    assert circ.evolve().memory[c[0]] == 1


def test_display_state() -> None:
    circ = qf.Circuit()
    circ += qf.X(1)
    circ += qf.X(2)
    circ += qf.StateDisplay(key="state0")

    ket = circ.run()
    assert "state0" in ket.memory

    rho = circ.evolve()
    assert "state0" in rho.memory


def test_display_probabilities() -> None:
    circ = qf.Circuit()
    circ += qf.X(1)
    circ += qf.X(2)
    circ += qf.ProbabilityDisplay(key="prob")

    ket = circ.run()
    assert "prob" in ket.memory

    rho = circ.evolve()
    assert "prob" in rho.memory


def test_density_display() -> None:
    circ = qf.Circuit()
    circ += qf.X(1)
    circ += qf.X(2)
    circ += qf.DensityDisplay(key="bloch1", qubits=[1])

    ket = circ.run()
    assert "bloch1" in ket.memory
    assert ket.memory["bloch1"].qubits == (1,)

    rho = circ.evolve()
    assert "bloch1" in rho.memory
    assert rho.memory["bloch1"].qubits == (1,)


def test_initialize() -> None:
    circ = qf.Circuit()
    circ += qf.H(1)
    ket = qf.random_state([0, 1, 2])
    circ += qf.Initialize(ket)

    assert circ.qubits == (0, 1, 2)
    assert qf.states_close(circ.run(), ket)

    assert qf.densities_close(circ.evolve(), ket.asdensity())


def test_reset() -> None:
    reset = qf.Reset(0, 1, 2)
    ket = qf.random_state([0, 1, 2, 3])
    ket = reset.run(ket)
    assert ket.tensor[1, 1, 0, 0] == 0.0
    assert str(reset) == "Reset 0 1 2"

    reset = qf.Reset()
    ket = qf.random_state([0, 1, 2, 3])
    ket = reset.run(ket)
    assert qf.states_close(ket, qf.zero_state([0, 1, 2, 3]))
    assert str(reset) == "Reset"

    with pytest.raises(TypeError):
        reset.evolve(qf.random_density([0, 1, 3]))

    with pytest.raises(TypeError):
        reset.asgate()

    with pytest.raises(TypeError):
        reset.aschannel()


# fin



#!/usr/bin/env python

# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""A collection of useful circuit identities"""

from itertools import zip_longest

import numpy as np
from sympy import Symbol, pi

import quantumflow as qf
from quantumflow import (
    CZ,
    XX,
    YY,
    ZZ,
    Can,
    CCNot,
    Circuit,
    CNot,
    CPhase,
    CPhase00,
    CPhase01,
    CPhase10,
    H,
    I,
    ISwap,
    PSwap,
    Rx,
    Rz,
    S,
    Swap,
    V,
    X,
    XPow,
    Y,
    YPow,
    Z,
    ZPow,
    circuit_to_diagram,
    gates_close,
    translate_ccnot_to_cnot,
)

syms = {
    "alpha": Symbol("α"),
    "lam": Symbol("λ"),
    "nx": Symbol("n_x"),
    "ny": Symbol("n_y"),
    "nz": Symbol("n_z"),
    "p": Symbol("p"),
    "phi": Symbol("φ"),
    "t": Symbol("t"),
    "t0": Symbol("t_0"),
    "t1": Symbol("t_1"),
    "t2": Symbol("t_2"),
    "theta": Symbol("θ"),
    "tx": Symbol("t_x"),
    "ty": Symbol("t_y"),
    "tz": Symbol("t_z"),
}


def identities():
    """Return a list of circuit identities, each consisting of a name, and
    two equivalent Circuits."""

    circuit_identities = []

    # Pick random parameter
    # theta = np.pi * np.random.uniform()
    # t = np.random.uniform()

    theta = syms["theta"]
    t = syms["t"]

    tx = syms["tx"]
    ty = syms["ty"]
    tz = syms["tz"]

    # Single qubit gate identities

    name = "Hadamard is own inverse"
    circ0 = Circuit([H(0), H(0)])
    circ1 = Circuit([I(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Hadamards convert X to Z"
    circ0 = Circuit([H(0), X(0), H(0)])
    circ1 = Circuit([Z(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Hadamards convert Z to X"
    circ0 = Circuit([H(0), Z(0), H(0)])
    circ1 = Circuit([X(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "S sandwich converts X to Y"
    circ0 = Circuit([S(0).H, X(0), S(0)])
    circ1 = Circuit([Y(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "S sandwich converts Y to X"
    circ0 = Circuit([S(0), Y(0), S(0).H])
    circ1 = Circuit([X(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Hadamards convert Rz to Rx"
    circ0 = Circuit([H(0), Rz(theta, 0), H(0)])
    circ1 = Circuit([Rx(theta, 0)])
    circuit_identities.append([name, circ0, circ1])

    # Simplified from Cirq's "_potential_cross_whole_w"
    # "_potential_cross_partial_w" is essentially same identity
    # X and ZPow commute
    name = "X-sandwich inverts ZPow"
    circ0 = Circuit([X(0), ZPow(t, 0), X(0)])
    circ1 = Circuit([ZPow(-t, 0)])
    circuit_identities.append([name, circ0, circ1])

    # Same relation  as previous
    name = "Commute X past ZPow"
    circ0 = Circuit([X(0), ZPow(t, 0)])
    circ1 = Circuit([ZPow(-t, 0), X(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute X past YPow"
    circ0 = Circuit([X(0), YPow(t, 0)])
    circ1 = Circuit([YPow(-t, 0), X(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute V past ZPow"
    circ0 = Circuit([V(0), ZPow(t, 0)])
    circ1 = Circuit([YPow(t, 0), V(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute V past YPow"
    circ0 = Circuit([V(0), YPow(t, 0)])
    circ1 = Circuit([ZPow(-t, 0), V(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute V.H past ZPow"
    circ0 = Circuit([V(0).H, ZPow(t, 0)])
    circ1 = Circuit([YPow(-t, 0), V(0).H])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute V.H past YPow"
    circ0 = Circuit([V(0).H, YPow(t, 0)])
    circ1 = Circuit([ZPow(t, 0), V(0).H])
    circuit_identities.append([name, circ0, circ1])

    name = "Couumte S.H past XPow"
    circ0 = Circuit([S(0).H, XPow(t, 0)])
    circ1 = Circuit([YPow(t, 0), S(0).H])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute S past XPow"
    circ0 = Circuit([S(0), XPow(t, 0)])
    circ1 = Circuit([YPow(-t, 0), S(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Y to ZX"
    circ0 = Circuit([Y(0)])
    circ1 = Circuit([Z(0), X(0)])
    circuit_identities.append([name, circ0, circ1])

    # ZYZ Decompositions

    name = "Hadamard ZYZ decomposition"
    circ0 = Circuit([H(0)])
    circ1 = Circuit([Z(0), Y(0) ** 0.5])
    circuit_identities.append([name, circ0, circ1])

    # CNot identities

    name = "CZ to CNot"
    circ0 = Circuit([CZ(0, 1)])
    circ1 = Circuit([H(1), CNot(0, 1), H(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "CZ to CNot (2)"
    circ0 = Circuit([CZ(0, 1)])
    circ1 = Circuit([YPow(+0.5, 1), CNot(0, 1), YPow(-0.5, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Swap to 3 CNots"
    circ0 = Circuit([Swap(0, 1)])
    circ1 = Circuit([CNot(0, 1), CNot(1, 0), CNot(0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Swap to 3 CZs"
    circ0 = Circuit([Swap(0, 1)])
    circ1 = Circuit([CZ(0, 1), H(0), H(1), CZ(1, 0), H(0), H(1), CZ(0, 1), H(0), H(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "ISwap decomposition to Swap and CNot"
    circ0 = Circuit([ISwap(0, 1)])
    circ1 = Circuit([Swap(0, 1), H(1), CNot(0, 1), H(1), S(0), S(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "ISwap decomposition to Swap and CZ"
    # This makes it clear why you can commute Rz's across ISwap
    circ0 = Circuit([ISwap(0, 1)])
    circ1 = Circuit([Swap(0, 1), CZ(0, 1), S(0), S(1)])
    circuit_identities.append([name, circ0, circ1])

    # http://info.phys.unm.edu/~caves/courses/qinfo-f14/lectures/lectures21-23.pdf
    name = "CNot sandwich with X on control"
    circ0 = Circuit([CNot(0, 1), X(0), CNot(0, 1)])
    circ1 = Circuit([X(0), X(1)])
    circuit_identities.append([name, circ0, circ1])

    # http://info.phys.unm.edu/~caves/courses/qinfo-f14/lectures/lectures21-23.pdf
    name = "CNot sandwich with Z on target"
    circ0 = Circuit([CNot(0, 1), Z(1), CNot(0, 1)])
    circ1 = Circuit([Z(0), Z(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "DCNot (Double-CNot) to iSwap"
    circ0 = Circuit([CNot(0, 1), CNot(1, 0)])
    circ1 = Circuit([H(0), S(0).H, S(1).H, ISwap(0, 1), H(1)])
    circuit_identities.append([name, circ0, circ1])

    # Commuting single qubit gates across 2 qubit games

    name = "Commute X on CNot target"
    circ0 = Circuit([X(1), CNot(0, 1)])
    circ1 = Circuit([CNot(0, 1), X(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute X on CNot control"
    circ0 = Circuit([X(0), CNot(0, 1)])
    circ1 = Circuit([CNot(0, 1), X(0), X(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute Z on CNot target"
    circ0 = Circuit([Z(1), CNot(0, 1)])
    circ1 = Circuit([CNot(0, 1), Z(0), Z(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute Z on CNot control"
    circ0 = Circuit([Z(0), CNot(0, 1)])
    circ1 = Circuit([CNot(0, 1), Z(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute X with CZ"
    circ0 = Circuit([X(0), CZ(0, 1)])
    circ1 = Circuit([CZ(0, 1), X(0), Z(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute X with XX"
    circ0 = Circuit([X(0), XX(t, 0, 1)])
    circ1 = Circuit(
        [
            XX(t, 0, 1),
            X(0),
        ]
    )
    circuit_identities.append([name, circ0, circ1])

    name = "Commute X with YY"
    circ0 = Circuit([X(0), YY(t, 0, 1)])
    circ1 = Circuit(
        [
            YY(-t, 0, 1),
            X(0),
        ]
    )
    circuit_identities.append([name, circ0, circ1])

    name = "Commute X with ZZ"
    circ0 = Circuit([X(0), ZZ(t, 0, 1)])
    circ1 = Circuit(
        [
            ZZ(-t, 0, 1),
            X(0),
        ]
    )
    circuit_identities.append([name, circ0, circ1])

    name = "Commute X with Canonical"
    circ0 = Circuit([X(0), Can(tx, ty, tz, 0, 1)])
    circ1 = Circuit([Can(tx, -ty, -tz, 0, 1), X(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute Y with Canonical"
    circ0 = Circuit([Y(0), Can(tx, ty, tz, 0, 1)])
    circ1 = Circuit([Can(-tx, ty, -tz, 0, 1), Y(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute Z with Canonical"
    circ0 = Circuit([Z(0), Can(tx, ty, tz, 0, 1)])
    circ1 = Circuit([Can(-tx, -ty, tz, 0, 1), Z(0)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commute Sqrt(X) with Canonical switches ty and tz arguments"
    circ0 = Circuit([V(0), V(1), Can(tx, ty, tz, 0, 1)])
    circ1 = Circuit([Can(tx, tz, ty, 0, 1), V(0), V(1)])
    circuit_identities.append([name, circ0, circ1])

    #  Canonical decompositions

    name = "Canonical gates: CZ to ZZ"
    circ0 = Circuit([CZ(0, 1), S(0), S(1)])
    circ1 = Circuit([ZZ(0.5, 0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Canonical gates: XX to ZZ"
    circ0 = Circuit([H(0), H(1), XX(0.5, 0, 1), H(0), H(1)])
    circ1 = Circuit([ZZ(0.5, 0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Canonical gates: CNot to XX"
    circ0 = Circuit([CNot(0, 1)])
    circ1 = Circuit([H(0), XX(0.5, 0, 1), H(0), H(1), S(0).H, S(1).H, H(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Canonical gates: Swap to Canonical"
    circ0 = Circuit([Swap(0, 1)])
    circ1 = Circuit([Can(0.5, 0.5, 0.5, 0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "ISwap to Canonical"
    circ0 = Circuit([ISwap(0, 1)])
    circ1 = Circuit([Can(-0.5, -0.5, 0.0, 0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "ISwap to Canonical in Weyl chamber"
    circ0 = Circuit([ISwap(0, 1)])
    circ1 = Circuit([X(0), Can(0.5, 0.5, 0.0, 0, 1), X(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "ISwap conjugate to ISwap"
    circ0 = Circuit([ISwap(0, 1).H])
    circ1 = Circuit([X(0), ISwap(0, 1), X(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "DCNot to Canonical"
    circ0 = Circuit([CNot(0, 1), CNot(1, 0)])
    circ1 = Circuit([H(0), S(0).H, S(1).H, X(0), Can(0.5, 0.5, 0.0, 0, 1), X(1), H(1)])
    circuit_identities.append([name, circ0, circ1])

    # Multi-qubit circuits

    name = "CNot controls commute"
    circ0 = Circuit([CNot(1, 0), CNot(1, 2)])
    circ1 = Circuit([CNot(1, 2), CNot(1, 0)])
    circuit_identities.append([name, circ0, circ1])

    name = "CNot targets commute"
    circ0 = Circuit([CNot(0, 1), CNot(2, 1)])
    circ1 = Circuit([CNot(2, 1), CNot(0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Commutation of CNot target/control"
    circ0 = Circuit([CNot(0, 1), CNot(1, 2)])
    circ1 = Circuit([CNot(1, 2), CNot(0, 2), CNot(0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Indirect CNot and 4 CNotS with intermediate qubit"
    circ0 = Circuit([CNot(0, 2), I(1)])
    circ1 = Circuit([CNot(0, 1), CNot(1, 2), CNot(0, 1), CNot(1, 2)])
    circuit_identities.append([name, circ0, circ1])

    name = "CZs with shared shared qubit commute"
    circ0 = Circuit([CZ(0, 1), CZ(1, 2)])
    circ1 = Circuit([CZ(1, 2), CZ(0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Z's shift ZZ by one unit"
    circ0 = Circuit([ZZ(t, 0, 1)])
    circ1 = Circuit([ZZ(t + 1, 0, 1), Z(0), Z(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "Z's shift ZZ by one unit"
    circ0 = Circuit([XX(t, 0, 1)])
    circ1 = Circuit([XX(t + 1, 0, 1), X(0), X(1)])
    circuit_identities.append([name, circ0, circ1])

    # Parametric circuits

    name = "ZZ to CNots"  # 1108.4318
    circ0 = Circuit([ZZ(theta / pi, 0, 1)])
    circ1 = Circuit([CNot(0, 1), Rz(theta, 1), CNot(0, 1)])
    circuit_identities.append([name, circ0, circ1])

    name = "XX to CNots"  # 1108.4318
    circ0 = Circuit([XX(theta / pi, 0, 1)])
    circ1 = Circuit([H(0), H(1), CNot(0, 1), Rz(theta, 1), CNot(0, 1), H(0), H(1)])
    circuit_identities.append([name, circ0, circ1])

    name = "XX to CNots (2)"
    circ0 = Circuit([XX(theta / pi, 0, 1)])
    circ1 = Circuit(
        [
            Y(0) ** 0.5,
            Y(1) ** 0.5,
            CNot(0, 1),
            Rz(theta, 1),
            CNot(0, 1),
            Y(0) ** -0.5,
            Y(1) ** -0.5,
        ]
    )
    circuit_identities.append([name, circ0, circ1])

    name = "YY to CNots"
    circ0 = Circuit([YY(theta / pi, 0, 1)])
    circ1 = Circuit(
        [
            X(0) ** 0.5,
            X(1) ** 0.5,
            CNot(0, 1),
            Rz(theta, 1),
            CNot(0, 1),
            X(0) ** -0.5,
            X(1) ** -0.5,
        ]
    )
    circuit_identities.append([name, circ0, circ1])

    def cphase_to_zz(gate: CPhase):
        t = -gate.parameter("theta") / (2 * pi)
        q0, q1 = gate.qubits
        circ = Circuit([ZZ(t, q0, q1), ZPow(-t, q0), ZPow(-t, q1)])
        return circ

    def cphase00_to_zz(gate: CPhase00):
        t = -gate.parameter("theta") / (2 * pi)
        q0, q1 = gate.qubits
        circ = Circuit(
            [X(0), X(1), ZZ(t, q0, q1), ZPow(-t, q0), ZPow(-t, q1), X(0), X(1)]
        )
        return circ

    def cphase01_to_zz(gate: CPhase00):
        t = -gate.parameter("theta") / (2 * pi)
        q0, q1 = gate.qubits
        circ = Circuit([X(0), ZZ(t, q0, q1), ZPow(-t, q0), ZPow(-t, q1), X(0)])
        return circ

    def cphase10_to_zz(gate: CPhase00):
        t = -gate.parameter("theta") / (2 * pi)
        q0, q1 = gate.qubits
        circ = Circuit([X(1), ZZ(t, q0, q1), ZPow(-t, q0), ZPow(-t, q1), X(1)])
        return circ

    name = "CPhase to ZZ"
    gate = CPhase(theta, 0, 1)
    circ0 = Circuit([gate])
    circ1 = cphase_to_zz(gate)
    circuit_identities.append([name, circ0, circ1])

    name = "CPhase00 to ZZ"
    gate = CPhase00(theta, 0, 1)
    circ0 = Circuit([gate])
    circ1 = cphase00_to_zz(gate)
    circuit_identities.append([name, circ0, circ1])

    name = "CPhase01 to ZZ"
    gate = CPhase01(theta, 0, 1)
    circ0 = Circuit([gate])
    circ1 = cphase01_to_zz(gate)
    circuit_identities.append([name, circ0, circ1])

    name = "CPhase10 to ZZ"
    gate = CPhase10(theta, 0, 1)
    circ0 = Circuit([gate])
    circ1 = cphase10_to_zz(gate)
    circuit_identities.append([name, circ0, circ1])

    name = "PSwap to Canonical"
    gate = PSwap(theta, 0, 1)
    circ0 = Circuit([gate])
    circ1 = Circuit()
    circ1 += YPow(1, 0)
    circ1 += Can(0.5, 0.5, 0.5 - theta / pi, 0, 1)
    circ1 += YPow(1, 1)
    circuit_identities.append([name, circ0, circ1])

    # Three qubit gates

    name = "Toffoli gate CNot decomposition"
    circ0 = Circuit([CCNot(0, 1, 2)])
    circ1 = translate_ccnot_to_cnot(CCNot(0, 1, 2))
    circuit_identities.append([name, circ0, circ1])

    name = "CCZ to CNots, respecting adjacency"
    gate = qf.CCZ(0, 1, 2)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_ccz_to_adjacent_cnot(gate))
    circuit_identities.append([name, circ0, circ1])

    name = "CCNot to CCZ"
    gate = qf.CCNot(0, 1, 2)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_ccnot_to_ccz(gate))
    circuit_identities.append([name, circ0, circ1])

    name = "CSwap to CCNot"
    gate = qf.CSwap(0, 1, 2)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_cswap_to_ccnot(gate))
    circuit_identities.append([name, circ0, circ1])

    name = "CSwap to CNot"
    gate = qf.CSwap(0, 1, 2)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_cswap_to_cnot(gate))
    circuit_identities.append([name, circ0, circ1])

    name = "CSwap to CNot (control between targets)"
    gate = qf.CSwap(1, 0, 2)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_cswap_inside_to_cnot(gate))
    circuit_identities.append([name, circ0, circ1])

    name = "CH to Clifford+T"
    gate = qf.CH(0, 1)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_ch_to_cpt(gate))
    circuit_identities.append([name, circ0, circ1])

    name = "CV to Clifford+T"
    gate = qf.CV(0, 1)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_cv_to_cpt(gate))
    circuit_identities.append([name, circ0, circ1])

    name = "CV_H to Clifford+T"
    gate = qf.CV_H(0, 1)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_cvh_to_cpt(gate))
    circuit_identities.append([name, circ0, circ1])

    # from sympy import Symbol, pi
    name = "CNotPow to ZZ"
    gate = qf.CNotPow(t, 0, 1)
    circ0 = Circuit([gate])
    circ1 = Circuit(qf.translate_cnotpow_to_zz(gate))
    circuit_identities.append([name, circ0, circ1])

    # name = "PISwap to XY"
    # gate = qf.PISwap(theta, 0, 1)
    # circ0 = Circuit([gate])
    # circ1 = Circuit(qf.translate_piswap_to_xy(gate))
    # circuit_identities.append([name, circ0, circ1])

    name = "Powers of iSwap to CNot sandwich"
    gate = qf.ISwap(0, 1) ** t
    trans = [qf.translate_xy_to_can, qf.translate_can_to_cnot]
    circ0 = Circuit([gate])
    circ1 = qf.circuit_translate(circ0, trans)
    circuit_identities.append([name, circ0, circ1])

    name = "Commute ZPow past multiple CNots"
    gate = qf.ZPow(t, 1)
    circ0 = Circuit([gate, qf.CNot(0, 1), qf.CNot(1, 2), qf.CNot(0, 1)])
    circ1 = Circuit([qf.CNot(0, 1), qf.CNot(1, 2), qf.CNot(0, 1), gate])
    circuit_identities.append([name, circ0, circ1])

    return circuit_identities


def _print_circuit_identity(
    name, circ0, circ1, min_col_width=0, col_sep=5, left_margin=8
):
    print("# ", name)

    circ0 = Circuit(circ0)
    circ1 = Circuit(circ1)

    gates0 = circuit_to_diagram(circ0, qubit_labels=False).splitlines()
    gates1 = circuit_to_diagram(circ1, qubit_labels=False).splitlines()

    for gate0, gate1 in zip_longest(gates0, gates1, fillvalue=""):
        line = (" " * col_sep).join(
            [gate0.ljust(min_col_width), gate1.ljust(min_col_width)]
        )
        line = (" " * left_margin) + line
        line = line.rstrip()
        print(line)

    print()
    print()

    concrete = {name: np.random.uniform() for name in syms.values()}
    circ0f = circ0.resolve(concrete)
    circ1f = circ1.resolve(concrete)
    assert gates_close(circ0f.asgate(), circ1f.asgate())


def _check_circuit_identities(circuit_identities):
    for name, circ0, circ1 in circuit_identities:
        _print_circuit_identity(name, circ0, circ1)


if __name__ == "__main__":
    print()
    print("# Validate and report various circuit identities " "(up to global phase)")
    print()
    _check_circuit_identities(identities())



# Copyright 2019-, Gavin E. Crooks and the QuantumFlow contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""Unit tests for quantumflow.circuits"""


import networkx as nx
import numpy as np
import pytest

import quantumflow as qf
from quantumflow.utils import bitlist_to_int, int_to_bitlist


def true_ket() -> qf.State:
    # Adapted from referenceQVM
    wf_true = np.array(
        [
            0.00167784 + 1.00210180e-05 * 1j,
            0.50000000 - 4.99997185e-01 * 1j,
            0.50000000 - 4.99997185e-01 * 1j,
            0.00167784 + 1.00210180e-05 * 1j,
        ]
    )
    return qf.State(wf_true.reshape((2, 2)))


def test_str() -> None:
    circ = qf.zyz_circuit(0.1, 2.2, 0.5, 0)
    str(circ)
    # TODO Expand


def test_name() -> None:
    assert qf.Circuit().name == "Circuit"


def test_qaoa_circuit() -> None:
    circ = qf.Circuit()
    circ += qf.Ry(np.pi / 2, 0)
    circ += qf.Rx(np.pi, 0)
    circ += qf.Ry(np.pi / 2, 1)
    circ += qf.Rx(np.pi, 1)
    circ += qf.CNot(0, 1)
    circ += qf.Rx(-np.pi / 2, 1)
    circ += qf.Ry(4.71572463191, 1)
    circ += qf.Rx(np.pi / 2, 1)
    circ += qf.CNot(0, 1)
    circ += qf.Rx(-2 * 2.74973750579, 0)
    circ += qf.Rx(-2 * 2.74973750579, 1)

    ket = qf.zero_state(2)
    ket = circ.run(ket)

    assert qf.states_close(ket, true_ket())


def test_qaoa_circuit_turns() -> None:
    circ = qf.Circuit()
    circ += qf.YPow(1 / 2, 0)
    circ += qf.XPow(1, 0)
    circ += qf.YPow(1 / 2, 1)
    circ += qf.XPow(1, 1)
    circ += qf.CNot(0, 1)
    circ += qf.XPow(-1 / 2, 1)
    circ += qf.YPow(4.71572463191 / np.pi, 1)
    circ += qf.XPow(1 / 2, 1)
    circ += qf.CNot(0, 1)
    circ += qf.XPow(-2 * 2.74973750579 / np.pi, 0)
    circ += qf.XPow(-2 * 2.74973750579 / np.pi, 1)

    ket = qf.zero_state(2)
    ket = circ.run(ket)

    assert qf.states_close(ket, true_ket())


def test_circuit_wires() -> None:
    circ = qf.Circuit()
    circ += qf.YPow(1 / 2, 0)
    circ += qf.XPow(1, 10)
    circ += qf.YPow(1 / 2, 1)
    circ += qf.XPow(1, 1)
    circ += qf.CNot(0, 4)

    bits = circ.qubits
    assert bits == (0, 1, 4, 10)


def test_inverse() -> None:
    # Random circuit
    circ = qf.Circuit()
    circ += qf.YPow(1 / 2, 0)
    circ += qf.H(0)
    circ += qf.YPow(1 / 2, 1)
    circ += qf.XPow(1.23123, 1)
    circ += qf.CNot(0, 1)
    circ += qf.XPow(-1 / 2, 1)
    circ += qf.YPow(4.71572463191 / np.pi, 1)
    circ += qf.CNot(0, 1)
    circ += qf.XPow(-2 * 2.74973750579 / np.pi, 0)
    circ += qf.XPow(-2 * 2.74973750579 / np.pi, 1)

    circ_inv = circ.H

    ket = circ.run()
    # qf.print_state(ket)

    ket = circ_inv.run(ket)
    # qf.print_state(ket)

    # print(ket.qubits)
    # print(true_ket().qubits)
    assert qf.states_close(ket, qf.zero_state(2))

    ket = qf.zero_state(2)
    circ += circ_inv
    ket = circ.run(ket)
    assert qf.states_close(ket, qf.zero_state(2))


def test_implicit_state() -> None:
    circ = qf.Circuit()
    circ += qf.YPow(1 / 2, 0)
    circ += qf.H(0)
    circ += qf.YPow(1 / 2, 1)

    ket = circ.run()  # Implicit state
    assert len(ket.qubits) == 2

    with pytest.raises(TypeError):
        # Should fail because qubits aren't sortable, so no standard ordering
        circ += qf.YPow(1 / 2, "namedqubit")
        circ.qubits


def test_elements() -> None:
    circ = qf.Circuit()
    circ1 = qf.Circuit()
    circ2 = qf.Circuit()
    circ1 += qf.Ry(np.pi / 2, 0)
    circ1 += qf.Rx(np.pi, 0)
    circ1 += qf.Ry(np.pi / 2, 1)
    circ1 += qf.Rx(np.pi, 1)
    circ1 += qf.CNot(0, 1)
    circ2 += qf.Rx(-np.pi / 2, 1)
    circ2 += qf.Ry(4.71572463191, 1)
    circ2 += qf.Rx(np.pi / 2, 1)
    circ2 += qf.CNot(0, 1)
    circ2 += qf.Rx(-2 * 2.74973750579, 0)
    circ2 += qf.Rx(-2 * 2.74973750579, 1)
    circ += circ1
    circ += circ2

    assert len(circ) == 11
    assert circ.size() == 11
    assert circ[4].name == "CNot"

    circ_13 = circ[1:3]
    assert len(circ_13) == 2
    assert isinstance(circ_13, qf.Circuit)


def test_create() -> None:
    gen = [qf.H(i) for i in range(8)]

    circ1 = qf.Circuit(list(gen))
    circ1.run(qf.zero_state(8))

    circ2 = qf.Circuit(gen)
    circ2.run(qf.zero_state(8))

    circ3 = qf.Circuit(qf.H(i) for i in range(8))
    circ3.run(qf.zero_state(8))


def test_add() -> None:
    circ = qf.Circuit()
    circ += qf.H(0)
    circ += qf.H(1)
    circ += qf.H(1)

    assert len(list(circ)) == 3

    circ = circ + circ
    assert len(circ) == 6

    circ += circ
    assert len(circ) == 12


def test_ccnot_circuit_evolve() -> None:
    rho0 = qf.random_state(3).asdensity()
    gate = qf.CCNot(0, 1, 2)
    circ = qf.Circuit(qf.translate_ccnot_to_cnot(gate))
    rho1 = gate.evolve(rho0)
    rho2 = circ.evolve(rho0)
    assert qf.densities_close(rho1, rho2)


def test_circuit_aschannel() -> None:
    rho0 = qf.random_state(3).asdensity()
    rho1 = qf.CCNot(0, 1, 2).evolve(rho0)
    gate = qf.CCNot(0, 1, 2)
    circ = qf.Circuit(qf.translate_ccnot_to_cnot(gate))
    chan = circ.aschannel()
    rho2 = chan.evolve(rho0)

    assert qf.densities_close(rho1, rho2)


def test_control_circuit() -> None:
    ccnot = qf.control_circuit([0, 1], qf.X(2))
    ket0 = qf.random_state(3)
    ket1 = qf.CCNot(0, 1, 2).run(ket0)
    ket2 = ccnot.run(ket0)
    assert qf.states_close(ket1, ket2)


def test_phase_estimation_circuit_1() -> None:
    N = 8
    phase = 1 / 4
    gate = qf.Rz(-4 * np.pi * phase, N)
    circ = qf.phase_estimation_circuit(gate, range(N))
    res = circ.run().measure()[0:N]
    est_phase = bitlist_to_int(res) / 2 ** N
    assert np.isclose(phase, est_phase)


def test_phase_estimation_circuit_2() -> None:
    N = 8
    phase = 12 / 256
    gate = qf.Rz(-4 * np.pi * phase, N)
    circ = qf.phase_estimation_circuit(gate, range(N))
    res = circ.run().measure()[0:N]
    est_phase = bitlist_to_int(res) / 2 ** N
    assert np.isclose(phase, est_phase)


def test_phase_estimation_circuit_3() -> None:
    N = 8
    phase = 12 / 256
    gate = qf.ZZ(-4 * phase, N, N + 1)
    circ = qf.phase_estimation_circuit(gate, range(N))
    res = circ.run().measure()[0:N]
    est_phase = bitlist_to_int(res) / 2 ** N
    assert np.isclose(phase, est_phase)

    with pytest.raises(ValueError):
        # Gate and output qubits overlap
        circ = qf.phase_estimation_circuit(gate, range(N + 1))


def test_addition_circuit() -> None:
    # Two bit addition circuit
    circ = qf.addition_circuit([0, 1], [2, 3], [4, 5])

    for c0 in range(0, 2):
        for a0 in range(0, 4):
            for a1 in range(0, 4):
                expected = a0 + a1 + c0
                b0 = int_to_bitlist(a0, 2)
                b1 = int_to_bitlist(a1, 2)
                bc = [c0]
                bits = tuple(b0 + b1 + bc + [0])

                state = np.zeros(shape=[2] * 6)
                state[bits] = 1
                ket = qf.State(state)

                ket = circ.run(ket)
                bits = ket.measure()
                res = bits[[5, 2, 3]]  # type: ignore
                res = bitlist_to_int(res)

                assert res == expected

    # Three bit addition circuit
    circ = qf.addition_circuit([0, 1, 2], [3, 4, 5], [6, 7])
    for c0 in range(0, 2):
        for a0 in range(0, 8):
            for a1 in range(0, 8):
                expected = a0 + a1 + c0
                b0 = int_to_bitlist(a0, 3)
                b1 = int_to_bitlist(a1, 3)
                bc = [c0]
                bits = tuple(b0 + b1 + bc + [0])

                state = np.zeros(shape=[2] * 8)
                state[bits] = 1
                ket = qf.State(state)

                ket = circ.run(ket)
                bits = ket.measure()
                res = bits[[7, 3, 4, 5]]  # type: ignore
                res = bitlist_to_int(res)

                # print(c0, a0, a1, expected, res)
                assert res == expected

    with pytest.raises(ValueError):
        qf.addition_circuit([0, 1, 2], [3, 4, 5, 6], [7, 8])

    with pytest.raises(ValueError):
        qf.addition_circuit([0, 1, 2], [3, 4, 5], [6, 7, 8])


def test_ghz_circuit() -> None:
    N = 12
    qubits = list(range(N))
    circ = qf.ghz_circuit(qubits)
    circ.run()


def test_map_gate() -> None:
    circ = qf.map_gate(qf.X(0), [[0], [1], [2]])
    assert circ[1].qubits[0] == 1

    circ = qf.map_gate(qf.CNot(0, 1), [[0, 1], [1, 2]])
    assert circ[1].qubits == (1, 2)


def test_count_operations() -> None:
    circ = qf.Circuit()
    circ += qf.H(0)
    circ += qf.H(1)
    circ += qf.H(1)
    op_count = qf.count_operations(circ)
    assert op_count == {qf.H: 3}

    circ = qf.addition_circuit([0, 1, 2], [3, 4, 5], [6, 7])
    op_count = qf.count_operations(circ)
    assert op_count == {qf.CNot: 13, qf.CCNot: 6}


def test_graph_circuit() -> None:
    graph = nx.grid_graph([2, 3])
    layers = 8

    params = qf.graph_circuit_params(graph, layers)
    _ = qf.graph_circuit(graph, layers, params)


def test_graph_state_circuit() -> None:
    graph = nx.grid_graph([3, 3])
    _ = qf.graph_state_circuit(graph)


def test_circuit_flat() -> None:
    circ0 = qf.Circuit([qf.X(0), qf.X(1)])
    circ1 = qf.Circuit([qf.Y(0), qf.Y(1)])
    circ2 = qf.Circuit([circ1, qf.Z(0), qf.Z(1)])
    circ = qf.Circuit([circ0, circ2])

    flat = qf.Circuit(circ.flat())
    assert len(flat) == 6
    assert flat[2].name == "Y"


def test_circuit_params() -> None:
    circ = qf.Circuit()
    circ += qf.X(0) ** 0.3
    circ += qf.Swap(1, 2)

    assert len(circ.params) == 1
    assert circ.params == (0.3,)

    with pytest.raises(ValueError):
        _ = circ.param("theta")



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Standard one qubit gates
"""

from typing import Dict, List, Type

import numpy as np

from .. import tensors, utils, var
from ..config import CONJ, SQRT
from ..ops import StdGate
from ..paulialgebra import Pauli, sI, sX, sY, sZ
from ..qubits import Qubit
from ..states import Density, State
from ..tensors import QubitTensor
from ..var import PI, Variable

__all__ = (
    "I",
    "Ph",
    "X",
    "Y",
    "Z",
    "H",
    "S",
    "T",
    "PhaseShift",
    "Rx",
    "Ry",
    "Rz",
    "Rn",
    "XPow",
    "YPow",
    "ZPow",
    "HPow",
    "S_H",
    "T_H",
    "V",
    "V_H",
    "SqrtY",
    "SqrtY_H",
)


def _specialize_gate(
    gate: StdGate, periods: List[float], opts: Dict[float, Type[StdGate]]
) -> StdGate:
    """Return a specialized instance of a given general gate. Used
    by the specialize code of various gates.

    Args:
        gate:       The gate instance to specialize
        periods:    The periods of the gate's parameters. Gate parameters
                    are wrapped to range[0,period]
        opts:       A map from particular gate parameters to a special case
                    of the original gate type
    """
    # params = list(gate.params)

    params = list(gate.params)

    # for p in params:
    #     if variable_is_symbolic(p):
    #         return gate

    params = [p % pd for p, pd in zip(params, periods)]

    for values, gatetype in opts.items():
        if np.isclose(params, values):
            return gatetype(*gate.qubits)  # type: ignore

    return type(gate)(*params, *gate.qubits)  # type: ignore


# Standard 1 qubit gates


class I(StdGate):  # noqa: E742
    r"""
    The 1-qubit identity gate.

    .. math::
        I() \equiv \begin{pmatrix} 1 & 0 \\ 0 & 1 \end{pmatrix}
    """
    cv_hermitian = True
    cv_tensor_structure = "identity"

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        return Pauli.zero()

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return tensors.asqutensor(np.eye(2))

    @property
    def H(self) -> "I":
        return self  # Hermitian

    def __pow__(self, t: Variable) -> "I":
        return self

    def run(self, ket: State) -> State:
        return ket

    def evolve(self, rho: Density) -> Density:
        return rho


class Ph(StdGate):
    r"""
    Apply a global phase shift of exp(i phi).

    Since this gate applies a global phase it technically doesn't need to
    specify qubits at all. But we instead anchor the gate to 1 specific
    qubit so that we can keep track of the phase as when manipulate gates,
    circuits, and DAGCircuits.

    We generally don't actually care about the global phase, since it has no
    physical meaning, although it does matter when constructing controlled gates.

    .. math::
        \operatorname{Ph}(\phi) \equiv \begin{pmatrix} e^{i \phi}& 0 \\
                              0 & e^{i \phi} \end{pmatrix}
    """
    # TODO
    # Ref: Explorations in Quantum Computing, Williams, p77
    # Ref: Barenco

    cv_tensor_structure = "diagonal"
    _diagram_labels = ["Ph({phi})"]

    def __init__(self, phi: float, q0: Qubit) -> None:
        super().__init__(params=[phi], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        (phi,) = self.params
        return -phi * sI(q0)

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        phi = var.asfloat(self.param("phi"))
        unitary = [[np.exp(1j * phi), 0.0], [0.0, np.exp(1j * phi)]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Ph":
        return self ** -1

    def __pow__(self, t: Variable) -> "Ph":
        return Ph(t * self.param("phi"), *self.qubits)

    def run(self, ket: State) -> State:
        (phi,) = self.params
        tensor = ket.tensor * np.exp(1j * phi)
        return State(tensor, ket.qubits, ket.memory)

    def evolve(self, rho: Density) -> Density:
        """Global phase shifts have no effect on density matrices. Returns argument
        unchanged."""
        return rho


# End class Ph


class X(StdGate):
    r"""
    A 1-qubit Pauli-X gate.

    .. math::
        X() &\equiv \begin{pmatrix} 0 & 1 \\ 1 & 0 \end{pmatrix}
    """
    cv_hermitian = True
    cv_tensor_structure = "permutation"

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return -(PI / 2) * (1 - sX(q0))

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = [[0, 1], [1, 0]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "X":
        return self  # Hermitian

    def __pow__(self, t: Variable) -> "XPow":
        return XPow(t, *self.qubits)

    def run(self, ket: State) -> State:
        (idx,) = ket.qubit_indices(self.qubits)
        take = utils.multi_slice(axes=[idx], items=[[1, 0]])
        tensor = ket.tensor[take]
        return State(tensor, ket.qubits, ket.memory)


# end class X


class Y(StdGate):
    r"""
    A 1-qubit Pauli-Y gate.

    .. math::
        Y() &\equiv \begin{pmatrix} 0 & -i \\ i & 0 \end{pmatrix}

    mnemonic: "Minus eye high".
    """
    cv_hermitian = True
    cv_tensor_structure = "monomial"

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return -(PI / 2) * (1 - sY(q0))

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = np.asarray([[0, -1.0j], [1.0j, 0]])
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Y":
        return self  # Hermitian

    def __pow__(self, t: Variable) -> "YPow":
        return YPow(t, *self.qubits)

    def run(self, ket: State) -> State:
        # This is fast Since X and Z have fast optimizations.
        ket = Z(*self.qubits).run(ket)
        ket = X(*self.qubits).run(ket)
        return ket


# end class Y


class Z(StdGate):
    r"""
    A 1-qubit Pauli-Z gate.

    .. math::
        Z() &\equiv \begin{pmatrix} 1 & 0 \\ 0 & -1 \end{pmatrix}
    """
    cv_hermitian = True
    cv_tensor_structure = "diagonal"

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return -(PI / 2) * (1 - sZ(q0))

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = np.asarray([[1, 0], [0, -1.0]])
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Z":
        return self  # Hermitian

    def __pow__(self, t: Variable) -> "ZPow":
        return ZPow(t, *self.qubits)

    def run(self, ket: State) -> State:
        return ZPow(1, *self.qubits).run(ket)


# end class Z


class H(StdGate):
    r"""
    A 1-qubit Hadamard gate.

    .. math::
        H() \equiv \frac{1}{\sqrt{2}}
        \begin{pmatrix} 1 & 1 \\ 1 & -1 \end{pmatrix}
    """
    cv_hermitian = True

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return (PI / 2) * ((sX(q0) + sZ(q0)) / np.sqrt(2) - 1)

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = np.asarray([[1, 1], [1, -1]]) / np.sqrt(2)
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "_H":  # See NB implementation note below
        return self  # Hermitian

    def __pow__(self, t: Variable) -> "HPow":
        return HPow(t, *self.qubits)

    def run(self, ket: State) -> State:
        axes = ket.qubit_indices(self.qubits)
        s0 = utils.multi_slice(axes, [0])
        s1 = utils.multi_slice(axes, [1])
        tensor = ket.tensor.copy()
        tensor[s1] -= tensor[s0]
        tensor[s1] *= -0.5
        tensor[s0] -= tensor[s1]
        tensor *= np.sqrt(2)
        return State(tensor, ket.qubits, ket.memory)


# Note: H().H -> H, but the method shadows the class, so we can't
# annotate directly.
_H = H

# End class H


class S(StdGate):
    r"""
    A 1-qubit phase S gate, equivalent to ``Z ** (1/2)``. The square root
    of the Z gate. Also sometimes denoted as the P gate.

    .. math::
        S() \equiv \begin{pmatrix} 1 & 0 \\ 0 & i \end{pmatrix}

    """
    cv_tensor_structure = "diagonal"

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return (PI / 2) * (sZ(q0) - 1) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = np.asarray([[1.0, 0.0], [0.0, 1.0j]])
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "S_H":
        return S_H(*self.qubits)

    def __pow__(self, t: Variable) -> "ZPow":
        return ZPow(t / 2, *self.qubits)

    def run(self, ket: State) -> State:
        return ZPow(1 / 2, *self.qubits).run(ket)


# end class S


class T(StdGate):
    r"""
    A 1-qubit T (pi/8) gate, equivalent to ``X ** (1/4)``. The forth root
    of the Z gate (up to global phase).

    .. math::
        \begin{pmatrix} 1 & 0 \\ 0 & e^{i \pi / 4} \end{pmatrix}
    """
    cv_tensor_structure = "diagonal"

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return (PI / 2) * (sZ(q0) - 1) / 4

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = [[1.0, 0.0], [0.0, np.exp(1j * np.pi / 4.0)]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "T_H":
        return T_H(*self.qubits)

    def __pow__(self, t: Variable) -> "ZPow":
        return ZPow(t / 4, *self.qubits)

    def run(self, ket: State) -> State:
        return ZPow(1 / 4, *self.qubits).run(ket)


# end class T


class PhaseShift(StdGate):
    r"""
    A 1-qubit parametric phase shift gate.
    Equivalent to Rz up to a global phase.

    .. math::
        \text{PhaseShift}(\theta) \equiv \begin{pmatrix}
         1 & 0 \\ 0 & e^{i \theta} \end{pmatrix}
    """
    cv_tensor_structure = "diagonal"

    def __init__(self, theta: Variable, q0: Qubit) -> None:
        super().__init__(params=[theta], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (theta,) = self.params
        (q0,) = self.qubits
        return theta * (sZ(q0) - 1) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        unitary = [[1.0, 0.0], [0.0, np.exp(1j * theta)]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "PhaseShift":
        return self ** -1

    def __pow__(self, t: Variable) -> "PhaseShift":
        return PhaseShift(self.param("theta") * t, *self.qubits)

    def run(self, ket: State) -> State:
        (theta,) = self.params
        return ZPow(theta / np.pi, *self.qubits).run(ket)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (theta,) = self.params
        t = theta / np.pi
        gate0 = ZPow(t, *qbs)
        gate1 = gate0.specialize()
        return gate1


# end class PhaseShift


class Rx(StdGate):
    r"""A 1-qubit Pauli-X parametric rotation gate.

    .. math::
        R_x(\theta) =   \begin{bmatrix*}[r]
                            \cos(\half\theta) & -i \sin(\half\theta) \\
                            -i \sin(\half\theta) & \cos(\half\theta)
                        \end{bmatrix*}

    Args:
        theta: Angle of rotation in Bloch sphere
    """
    _diagram_labels = ["Rx({theta})"]

    def __init__(self, theta: Variable, q0: Qubit) -> None:
        super().__init__(params=[theta], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (theta,) = self.params
        (q0,) = self.qubits
        return theta * sX(q0) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        unitary = [
            [np.cos(theta / 2), -1.0j * np.sin(theta / 2)],
            [-1.0j * np.sin(theta / 2), np.cos(theta / 2)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Rx":
        return self ** -1

    def __pow__(self, t: Variable) -> "Rx":
        return Rx(self.param("theta") * t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (theta,) = self.params
        t = theta / np.pi
        gate0 = XPow(t, *qbs)
        gate1 = gate0.specialize()
        return gate1


# end class Rx


class Ry(StdGate):
    r"""A 1-qubit Pauli-Y parametric rotation gate

    .. math::
        R_y(\theta) =   \begin{bmatrix*}[r]
            \cos(\half\theta) & -\sin(\half\theta)
            \\ \sin(\half\theta) & \cos(\half\theta)
                        \end{bmatrix*}
    Args:
        theta: Angle of rotation in Bloch sphere
    """

    _diagram_labels = ["Ry({theta})"]

    def __init__(self, theta: Variable, q0: Qubit) -> None:
        super().__init__(params=[theta], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (theta,) = self.params
        (q0,) = self.qubits
        return theta * sY(q0) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        unitary = [
            [np.cos(theta / 2.0), -np.sin(theta / 2.0)],
            [np.sin(theta / 2.0), np.cos(theta / 2.0)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Ry":
        return self ** -1

    def __pow__(self, t: Variable) -> "Ry":
        return Ry(self.param("theta") * t, *self.qubits)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (theta,) = self.params
        t = theta / np.pi
        gate0 = YPow(t, *qbs)
        gate1 = gate0.specialize()
        return gate1


# end class Ry


class Rz(StdGate):
    r"""A 1-qubit Pauli-X parametric rotation gate

    .. math::
        R_z(\theta) =   \begin{bmatrix*}
        e^{-i\half\theta} & 0 \\
        0 & e^{+i\half\theta}
                        \end{bmatrix*}

    Args:
        theta: Angle of rotation in Bloch sphere
    """
    cv_tensor_structure = "diagonal"
    _diagram_labels = ["Rz({theta})"]

    def __init__(self, theta: Variable, q0: Qubit) -> None:
        super().__init__(params=[theta], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return self.param("theta") * sZ(q0) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        unitary = [[np.exp(-theta * 0.5j), 0], [0, np.exp(theta * 0.5j)]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Rz":
        return self ** -1

    def __pow__(self, t: Variable) -> "Rz":
        return Rz(self.param("theta") * t, *self.qubits)

    def run(self, ket: State) -> State:
        (theta,) = self.params
        return ZPow(theta / np.pi, *self.qubits).run(ket)

    def specialize(self) -> StdGate:
        qbs = self.qubits
        (theta,) = self.params
        t = theta / np.pi
        gate0 = ZPow(t, *qbs)
        gate1 = gate0.specialize()
        return gate1


# end class Rz


# Other 1-qubit gates


class S_H(StdGate):
    r"""
    The inverse of the 1-qubit phase S gate, equivalent to
    ``Z ** -1/2``.

    .. math::
        \begin{pmatrix} 1 & 0 \\ 0 & -i \end{pmatrix}

    """
    cv_tensor_structure = "diagonal"
    _diagram_labels = ["S" + CONJ]

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return -PI * (sZ(q0) - 1) / 4

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = np.asarray([[1.0, 0.0], [0.0, -1.0j]])
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "S":
        return S(*self.qubits)

    def __pow__(self, t: Variable) -> "ZPow":
        return ZPow(-t / 2, *self.qubits)

    def run(self, ket: State) -> State:
        return ZPow(-1 / 2, *self.qubits).run(ket)


# end class S_H


class T_H(StdGate):
    r"""
    The inverse (complex conjugate) of the 1-qubit T (pi/8) gate, equivalent
    to ``Z ** -1/4``.

    .. math::
        \begin{pmatrix} 1 & 0 \\ 0 & e^{-i \pi / 4} \end{pmatrix}
    """
    cv_tensor_structure = "diagonal"
    _diagram_labels = ["T" + CONJ]

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return -PI * (sZ(q0) - 1) / 8

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        unitary = [[1.0, 0.0], [0.0, np.exp(-1j * np.pi / 4.0)]]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "T":
        return T(*self.qubits)

    def __pow__(self, t: Variable) -> "ZPow":
        return ZPow(-t / 4, *self.qubits)

    def run(self, ket: State) -> State:
        return ZPow(-1 / 4, *self.qubits).run(ket)


# end class T_H


class Rn(StdGate):
    r"""A 1-qubit rotation of angle theta about axis (nx, ny, nz)

    .. math::
        R_n(\theta) = \cos \frac{\theta}{2} I - i \sin\frac{\theta}{2}
            (n_x X+ n_y Y + n_z Z)

    Args:
        theta: Angle of rotation on Block sphere
        (nx, ny, nz): A three-dimensional real unit vector
    """
    _diagram_labels = ["Rn({theta}, {nx}, {ny}, {nz})"]

    def __init__(
        self, theta: Variable, nx: Variable, ny: Variable, nz: Variable, q0: Qubit
    ) -> None:

        norm = var.sqrt(nx ** 2 + ny ** 2 + nz ** 2)

        nx /= norm
        ny /= norm
        nz /= norm
        theta *= norm

        super().__init__(params=[theta, nx, ny, nz], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        theta, nx, ny, nz = self.params
        (q0,) = self.qubits
        return theta * (nx * sX(q0) + ny * sY(q0) + nz * sZ(q0)) / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = var.asfloat(self.param("theta"))
        nx = var.asfloat(self.param("nx"))
        ny = var.asfloat(self.param("ny"))
        nz = var.asfloat(self.param("nz"))

        cost = np.cos(theta / 2)
        sint = np.sin(theta / 2)
        unitary = [
            [cost - 1j * sint * nz, -1j * sint * nx - sint * ny],
            [-1j * sint * nx + sint * ny, cost + 1j * sint * nz],
        ]

        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "Rn":
        return self ** -1

    def __pow__(self, t: Variable) -> "Rn":
        theta, nx, ny, nz = self.params
        return Rn(t * theta, nx, ny, nz, *self.qubits)

    # TODO:     def specialize(self) -> Gate:


# end class RN


class XPow(StdGate):
    r"""Powers of the 1-qubit Pauli-X gate.

    .. math::
        XPow(t) = X^t = e^{i \pi t/2} R_X(\pi t)

    Args:
        t: Number of half turns (quarter cycles) on Block sphere
    """

    _diagram_labels = ["X^{t}"]

    def __init__(self, t: Variable, q0: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        (q0,) = self.qubits
        return t * (sX(q0) - 1) * PI / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = np.pi * var.asfloat(self.param("t"))
        phase = np.exp(0.5j * theta)
        unitary = [
            [phase * np.cos(theta / 2), phase * -1.0j * np.sin(theta / 2)],
            [phase * -1.0j * np.sin(theta / 2), phase * np.cos(theta / 2)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "XPow":
        return self ** -1

    def __pow__(self, t: Variable) -> "XPow":
        return XPow(t * self.param("t"), *self.qubits)

    def specialize(self) -> StdGate:
        opts = {0.0: I, 0.5: V, 1.0: X, 1.5: V_H, 2.0: I}
        return _specialize_gate(self, [2], opts)


# end class XPow


class YPow(StdGate):
    r"""Powers of the 1-qubit Pauli-Y gate.

    The pseudo-Hadamard gate is YPow(3/2), and its inverse is YPow(1/2).

    .. math::
        YPow(t) = Y^t = e^{i \pi t/2} R_Y(\pi t)

    Args:
        t: Number of half turns (quarter cycles) on Block sphere

    """
    _diagram_labels = ["Y^{t}"]

    def __init__(self, t: Variable, q0: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        (q0,) = self.qubits
        return t * (sY(q0) - 1) * PI / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = np.pi * var.asfloat(self.param("t"))
        phase = np.exp(0.5j * theta)
        unitary = [
            [phase * np.cos(theta / 2.0), phase * -np.sin(theta / 2.0)],
            [phase * np.sin(theta / 2.0), phase * np.cos(theta / 2.0)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "YPow":
        return self ** -1

    def __pow__(self, t: Variable) -> "YPow":
        return YPow(t * self.param("t"), *self.qubits)

    def specialize(self) -> StdGate:
        opts = {0.0: I, 1.0: Y, 2.0: I}
        return _specialize_gate(self, [2], opts)


# end class YPow


class ZPow(StdGate):
    r"""Powers of the 1-qubit Pauli-Z gate.

    .. math::
        ZPow(t) = Z^t = e^{i \pi t/2} R_Z(\pi t)

    Args:
        t: Number of half turns (quarter cycles) on Block sphere
    """
    cv_tensor_structure = "diagonal"
    _diagram_labels = ["Z^{t}"]

    def __init__(self, t: Variable, q0: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (t,) = self.params
        (q0,) = self.qubits
        return t * (sZ(q0) - 1) * PI / 2

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = np.pi * var.asfloat(self.param("t"))
        phase = np.exp(0.5j * theta)
        unitary = [
            [phase * np.exp(-theta * 0.5j), 0],
            [0, phase * np.exp(theta * 0.5j)],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "ZPow":
        return ZPow(-self.param("t"), *self.qubits)

    def __pow__(self, t: Variable) -> "ZPow":
        return ZPow(t * self.param("t"), *self.qubits)

    def run(self, ket: State) -> State:
        (t,) = self.params
        axes = ket.qubit_indices(self.qubits)
        s1 = utils.multi_slice(axes, [1])
        tensor = ket.tensor.copy()
        tensor[s1] *= np.exp(+1.0j * np.pi * t)
        return State(tensor, ket.qubits, ket.memory)

    def specialize(self) -> StdGate:
        opts = {0.0: I, 0.25: T, 0.5: S, 1.0: Z, 1.5: S_H, 1.75: T_H, 2.0: I}
        return _specialize_gate(self, [2], opts)


# end class ZPow


class HPow(StdGate):
    r"""
    Powers of the 1-qubit Hadamard gate.

    .. math::
        HPow(t) = H^t = e^{i \pi t/2}
        \begin{pmatrix}
            \cos(\tfrac{t}{2}) + \tfrac{i}{\sqrt{2}}\sin(\tfrac{t}{2})) &
            \tfrac{i}{\sqrt{2}} \sin(\tfrac{t}{2}) \\
            \tfrac{i}{\sqrt{2}} \sin(\tfrac{t}{2}) &
            \cos(\tfrac{t}{2}) -\tfrac{i}{\sqrt{2}} \sin(\frac{t}{2})
        \end{pmatrix}
    """
    _diagram_labels = ("H^{t}",)

    def __init__(self, t: Variable, q0: Qubit) -> None:
        super().__init__(params=[t], qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        return H(*self.qubits).hamiltonian * self.param("t")

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        theta = np.pi * var.asfloat(self.param("t"))
        phase = np.exp(0.5j * theta)
        unitary = [
            [
                phase * np.cos(theta / 2)
                - (phase * 1.0j * np.sin(theta / 2)) / np.sqrt(2),
                -(phase * 1.0j * np.sin(theta / 2)) / np.sqrt(2),
            ],
            [
                -(phase * 1.0j * np.sin(theta / 2)) / np.sqrt(2),
                phase * np.cos(theta / 2)
                + (phase * 1.0j * np.sin(theta / 2)) / np.sqrt(2),
            ],
        ]
        return tensors.asqutensor(unitary)

    @property
    def H(self) -> "HPow":
        return self ** -1

    def __pow__(self, t: Variable) -> "HPow":
        return HPow(t * self.param("t"), *self.qubits)

    def specialize(self) -> StdGate:
        opts = {0.0: I, 1.0: H, 2.0: I}
        return _specialize_gate(self, [2], opts)


# end class HPow


class V(StdGate):
    r"""
    Principal square root of the X gate, X-PLUS-90 gate.
    """

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return (sX(q0) - 1) * PI / 4

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return XPow(0.5, *self.qubits).tensor

    @property
    def H(self) -> "V_H":
        return V_H(*self.qubits)

    def __pow__(self, t: Variable) -> "XPow":
        return XPow(0.5 * t, *self.qubits)


# end class V


class V_H(StdGate):
    r"""
    Complex conjugate of the V gate, X-MINUS-90 gate.
    """

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return -(sX(q0) - 1) * PI / 4

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return XPow(-0.5, *self.qubits).tensor

    @property
    def H(self) -> "V":
        return V(*self.qubits)

    def __pow__(self, t: Variable) -> "XPow":
        return XPow(-0.5 * t, *self.qubits)


# end class V_H


class SqrtY(StdGate):
    r"""
    Principal square root of the Y gate.
    """
    _diagram_labels = [SQRT + "Y"]

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return (sY(q0) - 1) * PI / 4

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return YPow(0.5, *self.qubits).tensor

    @property
    def H(self) -> "SqrtY_H":
        return SqrtY_H(*self.qubits)

    def __pow__(self, t: Variable) -> "YPow":
        return YPow(0.5 * t, *self.qubits)


# end class SqrtY


class SqrtY_H(StdGate):
    r"""
    Complex conjugate of the np.sqrtY gate.
    """
    _diagram_labels = [SQRT + "Y" + CONJ]

    def __init__(self, q0: Qubit) -> None:
        super().__init__(qubits=[q0])

    @property
    def hamiltonian(self) -> Pauli:
        (q0,) = self.qubits
        return -(sY(q0) - 1) * PI / 4

    @utils.cached_property
    def tensor(self) -> QubitTensor:
        return YPow(-0.5, *self.qubits).tensor

    @property
    def H(self) -> "SqrtY":
        return SqrtY(*self.qubits)

    def __pow__(self, t: Variable) -> "YPow":
        return YPow(-0.5 * t, *self.qubits)


# end class SqrtY_H


# fin



#!/usr/bin/env python

# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
QuantumFlow examples
"""

import quantumflow as qf


def prepare_w16():
    """
    Prepare a 16-qubit W state using sqrt(iswaps) and local gates,
    respecting linear topology

    """
    ket = qf.zero_state(16)
    circ = w16_circuit()
    ket = circ.run(ket)
    return ket


def w16_circuit() -> qf.Circuit:
    """
    Return a circuit that prepares the the 16-qubit W state using\
    sqrt(iswaps) and local gates, respecting linear topology
    """

    gates = [
        qf.X(7),
        qf.ISwap(7, 8) ** 0.5,
        qf.S(8),
        qf.Z(8),
        qf.Swap(7, 6),
        qf.Swap(6, 5),
        qf.Swap(5, 4),
        qf.Swap(8, 9),
        qf.Swap(9, 10),
        qf.Swap(10, 11),
        qf.ISwap(4, 3) ** 0.5,
        qf.S(3),
        qf.Z(3),
        qf.ISwap(11, 12) ** 0.5,
        qf.S(12),
        qf.Z(12),
        qf.Swap(3, 2),
        qf.Swap(4, 5),
        qf.Swap(11, 10),
        qf.Swap(12, 13),
        qf.ISwap(2, 1) ** 0.5,
        qf.S(1),
        qf.Z(1),
        qf.ISwap(5, 6) ** 0.5,
        qf.S(6),
        qf.Z(6),
        qf.ISwap(10, 9) ** 0.5,
        qf.S(9),
        qf.Z(9),
        qf.ISwap(13, 14) ** 0.5,
        qf.S(14),
        qf.Z(14),
        qf.ISwap(1, 0) ** 0.5,
        qf.S(0),
        qf.Z(0),
        qf.ISwap(2, 3) ** 0.5,
        qf.S(3),
        qf.Z(3),
        qf.ISwap(5, 4) ** 0.5,
        qf.S(4),
        qf.Z(4),
        qf.ISwap(6, 7) ** 0.5,
        qf.S(7),
        qf.Z(7),
        qf.ISwap(9, 8) ** 0.5,
        qf.S(8),
        qf.Z(8),
        qf.ISwap(10, 11) ** 0.5,
        qf.S(11),
        qf.Z(11),
        qf.ISwap(13, 12) ** 0.5,
        qf.S(12),
        qf.Z(12),
        qf.ISwap(14, 15) ** 0.5,
        qf.S(15),
        qf.Z(15),
    ]
    circ = qf.Circuit(gates)

    return circ


if __name__ == "__main__":

    def main():
        """CLI"""
        print(prepare_w16.__doc__)
        print("states           : probabilities")
        ket = prepare_w16()
        qf.print_probabilities(ket)

    main()



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

import random

import numpy as np

import quantumflow as qf

from ..config_test import REPS


def test_FSim() -> None:
    for _ in range(REPS):
        theta = random.uniform(-np.pi, +np.pi)
        phi = random.uniform(-np.pi, +np.pi)
        gate0 = qf.FSim(theta, phi, 0, 1)

        # Test with decomposition from Cirq.
        circ = qf.Circuit()
        circ += qf.XX(theta / np.pi, 0, 1)
        circ += qf.YY(theta / np.pi, 0, 1)
        circ += qf.CZ(0, 1) ** (-phi / np.pi)
        gate1 = circ.asgate()
        assert qf.gates_close(gate0, gate1)

        assert qf.gates_close(gate1.H, gate0.H)


def test_PhasedX() -> None:
    q0 = "4"
    for _ in range(REPS):
        t = random.uniform(0, 1)
        p = random.uniform(-2, +2)

        gate0 = qf.PhasedX(p, q0)
        gate1 = qf.PhasedXPow(p, t, q0)
        assert qf.gates_close(gate0 ** t, gate1)
        assert (gate0 ** t).qubits == (q0,)

        gate0.H
        gate1.H
        gate2 = gate1 ** t
        p2, t2 = gate2.params
        assert p2 == p
        assert np.isclose(t2 - t ** 2, 0.0)

        assert qf.gates_close(gate0, gate0.specialize())

    assert qf.gates_close(qf.PhasedX(-2.0, q0).specialize(), qf.X(q0))



# Copyright 2019-, Gavin E. Crooks and contributors
# Copyright 2016-2018, Rigetti Computing
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Unittests for QuantumFlow Gate Decompositions
"""

import numpy as np
import pytest
import scipy.stats
from numpy import pi

import quantumflow as qf
from quantumflow.decompositions import _eig_complex_symmetric

from .config_test import REPS


def test_bloch_decomposition() -> None:
    theta = 1.23

    gate0 = qf.Rn(theta, 1, 0, 0, 9)
    gate1 = qf.bloch_decomposition(gate0)[0]
    assert isinstance(gate1, qf.Gate)
    assert qf.gates_close(gate0, gate1)

    gate0 = qf.Rn(theta, 0, 1, 0, 9)
    gate1 = qf.bloch_decomposition(gate0)[0]
    assert isinstance(gate1, qf.Gate)
    assert qf.gates_close(gate0, gate1)

    gate0 = qf.Rn(theta, 0, 0, 1, 9)
    gate1 = qf.bloch_decomposition(gate0)[0]
    assert isinstance(gate1, qf.Gate)
    assert qf.gates_close(gate0, gate1)

    gate0 = qf.Rn(pi, np.sqrt(2), 0, np.sqrt(2), 9)
    gate1 = qf.bloch_decomposition(gate0)[0]
    assert isinstance(gate1, qf.Gate)
    assert qf.gates_close(gate0, gate1)

    for _ in range(REPS):
        gate3 = qf.RandomGate(qubits=[0])
        gate4 = qf.bloch_decomposition(gate3)[0]
        assert isinstance(gate4, qf.Gate)
        assert qf.gates_close(gate3, gate4)

    gate5 = qf.I(0)
    gate6 = qf.bloch_decomposition(gate5)[0]
    assert isinstance(gate6, qf.Gate)
    assert qf.gates_close(gate5, gate6)


def test_bloch_decomp_errors() -> None:
    # Wrong number of qubits
    with pytest.raises(ValueError):
        qf.bloch_decomposition(qf.CNot(0, 1))


def test_zyz_decomp_errors() -> None:
    # Wrong number of qubits
    with pytest.raises(ValueError):
        qf.zyz_decomposition(qf.CNot(0, 1))


def test_zyz_decomposition() -> None:
    gate0 = qf.RandomGate([1])
    circ1 = qf.zyz_decomposition(gate0)
    gate1 = circ1.asgate()
    assert qf.gates_close(gate0, gate1)


def test_euler_decomposition() -> None:
    gate0 = qf.RandomGate([1])

    for order in ["XYX", "XZX", "YXY", "YZY", "ZXZ", "ZYZ"]:
        circ1 = qf.euler_decomposition(gate0, euler=order)
        gate1 = circ1.asgate()
        assert qf.gates_close(gate0, gate1)


def test_kronecker_decomposition() -> None:
    for _ in range(REPS):
        left = qf.RandomGate([1]).asoperator()
        right = qf.RandomGate([1]).asoperator()
        both = np.kron(left, right)
        gate0 = qf.Unitary(both, [0, 1])
        circ = qf.kronecker_decomposition(gate0)
        gate1 = circ.asgate()

        assert qf.gates_close(gate0, gate1)

    circ2 = qf.Circuit()
    circ2 += qf.Z(0)
    circ2 += qf.H(1)
    gate2 = circ2.asgate()
    circ3 = qf.kronecker_decomposition(gate2)
    gate3 = circ3.asgate()
    assert qf.gates_close(gate2, gate3)

    circ4 = qf.kronecker_decomposition(gate0, euler="XYX")
    gate4 = circ4.asgate()
    assert qf.gates_close(gate0, gate4)


def test_kronecker_decomp_errors() -> None:
    # Wrong number of qubits
    with pytest.raises(ValueError):
        qf.kronecker_decomposition(qf.X(0))

    # Not kronecker product
    with pytest.raises(ValueError):
        qf.kronecker_decomposition(qf.CNot(0, 1))


def test_canonical_decomposition() -> None:
    for tt1 in range(0, 6):
        for tt2 in range(tt1):
            for tt3 in range(tt2):
                t1, t2, t3 = tt1 / 12, tt2 / 12, tt3 / 12
                if t3 == 0 and t1 > 0.5:
                    continue
                coords = np.asarray((t1, t2, t3))

                circ0 = qf.Circuit()
                circ0 += qf.RandomGate([0])
                circ0 += qf.RandomGate([1])
                circ0 += qf.Can(t1, t2, t3, 0, 1)
                circ0 += qf.RandomGate([0])
                circ0 += qf.RandomGate([1])
                gate0 = circ0.asgate()

                circ1 = qf.canonical_decomposition(gate0)
                assert qf.gates_close(gate0, circ1.asgate())

                canon = circ1[1]
                new_coords = np.asarray([canon.param(n) for n in ["tx", "ty", "tz"]])
                assert np.allclose(coords, np.asarray(new_coords))

                coords2 = qf.canonical_coords(gate0)
                assert np.allclose(coords, np.asarray(coords2))


def test_canonical_decomp_sandwich() -> None:
    for _ in range(REPS):
        # Random CZ sandwich
        circ0 = qf.Circuit()
        circ0 += qf.RandomGate([0])
        circ0 += qf.RandomGate([1])
        circ0 += qf.CZ(0, 1)
        circ0 += qf.YPow(0.4, 0)
        circ0 += qf.YPow(0.25, 1)
        circ0 += qf.CZ(0, 1)
        circ0 += qf.RandomGate([0])
        circ0 += qf.RandomGate([1])

        gate0 = circ0.asgate()

        circ1 = qf.canonical_decomposition(gate0)
        gate1 = circ1.asgate()

        assert qf.gates_close(gate0, gate1)
        assert qf.almost_unitary(gate0)


def test_canonical_decomp_random() -> None:
    for _ in range(REPS * 2):
        gate0 = qf.RandomGate([0, 1])
        gate1 = qf.canonical_decomposition(gate0).asgate()
        assert qf.gates_close(gate0, gate1)


def test_canonical_decomp_errors() -> None:
    # Wrong number of qubits
    with pytest.raises(ValueError):
        qf.canonical_decomposition(qf.X(0))


def test_decomp_stdgates() -> None:
    gate0 = qf.IdentityGate([0, 1])
    gate1 = qf.canonical_decomposition(gate0).asgate()
    assert qf.gates_close(gate0, gate1)

    gate2 = qf.CNot(0, 1)
    gate3 = qf.canonical_decomposition(gate2).asgate()
    assert qf.gates_close(gate2, gate3)

    gate4 = qf.Swap(0, 1)
    gate5 = qf.canonical_decomposition(gate4).asgate()
    assert qf.gates_close(gate4, gate5)

    gate6 = qf.ISwap(0, 1)
    gate7 = qf.canonical_decomposition(gate6).asgate()
    assert qf.gates_close(gate6, gate7)

    gate8 = qf.CNot(0, 1) ** 0.5
    gate9 = qf.canonical_decomposition(gate8).asgate()
    assert qf.gates_close(gate8, gate9)

    gate10 = qf.Swap(0, 1) ** 0.5
    gate11 = qf.canonical_decomposition(gate10).asgate()
    assert qf.gates_close(gate10, gate11)

    gate12 = qf.ISwap(0, 1) ** 0.5
    gate13 = qf.canonical_decomposition(gate12).asgate()
    assert qf.gates_close(gate12, gate13)


def test_decomp_sqrtswap_sandwich() -> None:
    circ0 = qf.Circuit()
    circ0 += qf.Can(1 / 4, 1 / 4, 1 / 4, 0, 1)
    circ0 += qf.RandomGate([0])
    circ0 += qf.RandomGate([1])
    circ0 += qf.Can(1 / 4, 1 / 4, 1 / 4, 0, 1)

    gate0 = circ0.asgate()
    circ1 = qf.canonical_decomposition(gate0)
    gate1 = circ1.asgate()
    assert qf.gates_close(gate0, gate1)


def test_eig_complex_symmetric() -> None:
    samples = 100
    for _ in range(samples):

        # Build a random symmetric complex matrix
        orthoganal = scipy.stats.special_ortho_group.rvs(4)
        eigvals = (
            np.random.normal(size=(4,)) + 1j * np.random.normal(size=(4,))
        ) / np.sqrt(2.0)
        M = orthoganal @ np.diag(eigvals) @ orthoganal.T

        eigvals, eigvecs = _eig_complex_symmetric(M)
        assert np.allclose(M, eigvecs @ np.diag(eigvals) @ eigvecs.T)


def test_eigcs_errors() -> None:
    with pytest.raises(np.linalg.LinAlgError):
        _eig_complex_symmetric(np.random.normal(size=(4, 4)))


def test_b_decomposition() -> None:
    for _ in range(REPS):
        gate0 = qf.RandomGate([4, 8])
        circ = qf.b_decomposition(gate0)
        gate1 = circ.asgate()
        assert qf.gates_close(gate0, gate1)


def test_cnot_decomposition() -> None:
    for _ in range(REPS):
        gate0 = qf.RandomGate([4, 8])
        circ = qf.cnot_decomposition(gate0)
        gate1 = circ.asgate()
        assert qf.gates_close(gate0, gate1)


def test_convert_can_to_weyl() -> None:
    for r in range(100):
        tx = np.random.uniform(-10, +10)
        ty = np.random.uniform(-10, +10)
        tz = np.random.uniform(-10, +10)

        gate0 = qf.Can(tx, ty, tz, 1, 2)
        circ = qf.convert_can_to_weyl(gate0, euler="XYX")

        assert qf.gates_close(gate0, circ.asgate())


# Fin



# Copyright 2020-, Gavin E. Crooks and contributors
#
# This source code is licensed under the Apache License, Version 2.0 found in
# the LICENSE.txt file in the root directory of this source tree.

"""
Package wide configuration
"""

import platform
import re
import sys
import typing

from .utils import importlib_metadata

__all__ = ["__version__", "about"]


package_name = "quantumflow"

try:
    __version__ = importlib_metadata.version(package_name)  # type: ignore
except Exception:  # pragma: no cover
    # package is not installed
    __version__ = "?.?.?"


# See https://numpy.org/doc/stable/reference/generated/numpy.allclose.html

RTOL = 1e-05
"""Default relative tolerance for numerical comparisons"""

ATOL = 1e-07
"""Default absolute tolerance for numerical comparisons"""


CIRCUIT_INDENT = 4
"""Number of spaces to indent when listing circuits"""


# --- Defaults for gate visualization as unicode text ---
# Used by visualizations.circuit_to_diagram()

TARGET = "X"
# TARGET = '⨁'              # "n-ary circled plus", U+2A01
SWAP_TARGET = "x"
# SWAP_TARGET = '×'         # Multiplication sign
XCTRL = "⊖"  # ⊖ 'circled minus'
NXCTRL = "⊕"  # ⊕ 'circled plus'
YCTRL = "⊘"  # ⊘ 'circled division slash'
NYCTRL = "⊗"  # ⊗ 'circled times'
CTRL = "●"
NCTRL = "○"
CONJ = "⁺"  # Unicode "superscript plus sign"
SQRT = "√"


def about(file: typing.TextIO = None) -> None:
    """Print information about the configuration

     ``> python -m quantumflow.about``

    Args:
        file: Output stream (Defaults to stdout)
    """
    name_width = 24
    versions = {}
    versions["platform"] = platform.platform(aliased=True)
    versions[package_name] = __version__
    versions["python"] = sys.version[0:5]

    for req in importlib_metadata.requires(package_name):  # type: ignore
        name = re.split("[; =><]", req)[0]
        try:
            versions[name] = importlib_metadata.version(name)  # type: ignore
        except Exception:  # pragma: no cover
            pass

    print(file=file)
    print("# Configuration (> python -m quantumflow.about)", file=file)
    for name, vers in versions.items():
        print(name.ljust(name_width), vers, file=file)
    print(file=file)


# fin
